
#include "globdef.h"
#include "uidef.h"
#include "sigdef.h"
#include "screendef.h"
#include "fft3def.h"
#include "options.h"
#include "thrdef.h"
#include "options.h"

// ************************************************************
// ************************************************************
// These defines allow a lot of information to be written to dmp
#if DUMPFILE == TRUE
#define PR00 0
#define PR01 0
#define PR02 0
#define PR03 0
#define PR04 0
#define PR05 0
#else
#define PR00 0
#define PR01 0
#define PR02 0
#define PR03 0
#define PR04 0
#define PR05 0
#endif

#define PRT00 if(PR00 != 0)DEB
#define PRT01 if(PR01 != 0)DEB
#define PRT02 if(PR02 != 0)DEB
#define PRT03 if(PR03 != 0)DEB
#define PRT04 if(PR04 != 0)DEB
#define PRT05 if(PR05 != 0)DEB
// ************************************************************
// ************************************************************

void show_cw(char *s);
void check_cw(int num,int type);

void make_coherent_graph(int clear_old);
void manage_meter_graph(void);
int coh_tophat_filter(float *xin,float *xout,int lk,int fpt);
int coherent_graph_scro;
int cg_old_y1;
int cg_old_y2;
int cg_old_x1;
int cg_old_x2;

extern int meter_graph_scro;
int zzpos;

// float[2] baseb_out=data to send to loudspeaker
// float[2] baseb=complex amplitude of first level coherent data.
// float[2] baseb_raw=complex amplitude of baseband signal. Raw data, pol. adapted.
// float[2] baseb_raw_orthog=complex amplitude of pol. orthog signal. Raw data.
// float[2] baseb_carrier=phase of carrier. Complex data, cosines and sines.
// float[1] baseb_carrier_ampl=amplitude of carrier
// float[1] baseb_totpwr=total power of baseb_raw
// float[2] baseb_envelope=complex amplitude from fitted dots and dashes.
// float[1] baseb_upthreshold=forward peak detector for power
// float[1] baseb_threshold=combined forward and backward peak detector at -6dB
// float[2] baseb_fit=fitted dots and dashes.
// float[2] baseb_tmp=array for intermediate data in complex format
// float[1] baseb_agc_level=used only when AGC is enabled.
// short int[1] baseb_ramp=indicator for time of power above/below threshold.
// short_int[1] baseb_clock=CW code clock
// float[2] baseb_tmp=for debug during development

// baseb_pa  baseb exists up to baseb_pa-1.
// baseb_pb  The point up to which thresholds exist.
// baseb_pc  The point up to which ramp is collected.
// baseb_pd  A key up region close before baseb_pc
// baseb_pe  The point up to which we have run first detect.
// baseb_pf
// baseb_px  The oldest point that contains valid data.

#define COH_GRAPH_ACTIVE -1000
#define X_SIZE 5

float evaluate_cw_spectrum(void)
{
int ia, ib, ic, kk, nn;
float t1, t2;
// cw_spectrum contains the modulation of the Morse-coded signal.
// One of the strong components should be the frequency of
// a string of Morse code dots, the "clock frequency" of the
// Morse code.
// The user has to set the baseband filter to allow this frequency
// within the baseband filter but the filter must not be set
// more than 2.5 times wider than optimum because we do not search
// for the "Morse clock" lower than this.
// The full baseband bandwidth, bw, corresponds to the last point
// in cw_spectrum. [cw_spectrum_size-1].
// Start at bw/3 and step upwards until a minimum is reached.
// Look for signal power in pairs of points.
ia=1+cw_spectrum_size/3;
t1=cw_spectrum[ia]+cw_spectrum[ia-1];
t2=cw_spectrum[ia]+cw_spectrum[ia+1];
while(t2<t1 && ia<cw_spectrum_size-2)
  {
  t1=t2;
  ia++;
  t2=cw_spectrum[ia]+cw_spectrum[ia+1];
  }
// Now we should have stepped away from the peak associated
// with the "Morse clock" divided by three.
// Look for the strongest peak above ia.
ib=ia;
t1=cw_spectrum[ib];
kk=ib;
while(ib<cw_spectrum_size)
  {
  if(t1<cw_spectrum[ib])
    {
    t1=cw_spectrum[ib];
    kk=ib;
    }
  ib++;
  }
// kk may be the Morse code clock, but it could also be
// half the frequency. Search 1.5 to 2.5 times kk for another
// maximum. If it has at least 50% of the power, attribute the
// higher peak to the Morse clock.
ib=1.5*kk;
ic=2.5*kk+1;
if(ic > cw_spectrum_size)ic=cw_spectrum_size;
nn=0;
t2=0;
while(ib < ic)
  {
  if(t2<cw_spectrum[ib])
    {
    t2=cw_spectrum[ib];
    nn=ib;
    }
  ib++;  
  }
if(2*t2 > t1)kk=nn;
// Get the peak posuition with some decimals. Not very accurate,
// but the first decimal should be close to correct.
if(kk < cw_spectrum_size-1)
  {
  parabolic_fit(&t2, &t1, sqrt(cw_spectrum[kk-1]),
                           sqrt(cw_spectrum[kk]),sqrt(cw_spectrum[kk+1]));
  t1+=kk;                           
  }
else 
  {
  t1=kk;
  t2=sqrt(cw_spectrum[kk]);
  } 
return t1;
}

void collect_ramp(void)
{
int ia, ib;
// Make a first decision key up/key down based on signal power.
// This is the conventional approach, a matched filter (the operator
// has to select the optimum bandwidth) followed by a squarelaw detector.
// The threshold is conventionally kept at a fixed level but here it is
// very close to the noise when no signal is present. while it is at
// half the peak amplitude for strong signals.
// Store the result in baseb_ramp as ramps.
// Positive values indicate key down while negative values
// indicate key up (or signal too weak)
ib=(baseb_pc+baseband_mask)&baseband_mask;
ia=baseb_pc;
while( ia != baseb_pb )
  {
  if(baseb_totpwr[ia] > baseb_threshold[ia])
    {
    if(baseb_ramp[ib]>0)
      {
      baseb_ramp[ia]=baseb_ramp[ib]+1;
      }
    else
      {
      baseb_ramp[ia]=1;
      }
    }
  else
    {
    if(baseb_ramp[ib]<0)
      {
      baseb_ramp[ia]=baseb_ramp[ib]-1;
      }
    else
      {
      baseb_ramp[ia]=-1;
      }
    }
  ib=ia;
  ia=(ia+1)&baseband_mask;
  }
baseb_pc=ib;
// Step baseb_pc backwards until a long region is found. A region of more than
// 5 units is likely to be key up, but if it is key down, back across it
// and continue to search for a long key-up.
while(abs(baseb_ramp[baseb_pc]) < 5*cg_code_unit || baseb_ramp[baseb_pc] > 0) 
  {
  baseb_pc=(baseb_pc-abs(baseb_ramp[baseb_pc])+baseband_size)&baseband_mask;
  if( ((baseb_pc-baseb_pe+baseband_size)&baseband_mask)>baseband_neg)
    {
    baseb_pc=baseb_pe;
    return;
    }
  }
// Now we know that baseb_pc points to the start of a long negative ramp.
}


void make_ideal_waveform(void)
{
int i, j, k, m, ia;
float t1;
// Set the size we use for averaging the waveform around dashes.
// Make it an even number.
cw_avg_points=15*cwbit_pts+1;
cw_avg_points&=0xfffffffe;
// Assuming the modulation spectrum has been correctly evaluated
// the length of a Morse code bit should be about cwbit_pts points.
// Construct a dash of the appropriate lengths, make it symmetric
// around mix2_size/2 and run it through the baseband filter.
// Store the result as the first guess for our CW waveforms.
for(i=0; i<2*mix2_size; i++) mix2_tmp[i]=0;
j=mix2_size;
t1=cwbit_pts*3.0;
k=t1;
m=k/2;
k=2*m;
mix2_tmp[j]=1; 
for(i=0; i<m; i++)
  {
  mix2_tmp[j+2*i]=1;
  mix2_tmp[j-2-2*i]=1;
  }
t1=0.5*(t1-k);
// t1 is the amount by which the dash is longer than the
// amount we set to 1 (on both sides).
// With full level for a fractional (=t1) sample period,
// the power within the interval is t1. 
// The amplitude we should set is therefore sqrt(t1)
mix2_tmp[j+2*m]=sqrt(t1);
mix2_tmp[j-2-2*m]=sqrt(t1);
fftforward(mix2_size, mix2_n, mix2_tmp, mix2_table, mix2_permute);
i=1;
j=fft3_size/2+1;
while(bg_filterfunc[j] != 0)
  {
  mix2_tmp[2*i  ]*=bg_filterfunc[j];    
  mix2_tmp[2*i+1]*=bg_filterfunc[j];    
  mix2_tmp[2*(mix2_size-i)  ]*=bg_filterfunc[j];    
  mix2_tmp[2*(mix2_size-i)+1]*=bg_filterfunc[j];    
  i++;
  j++;
  }
while(i <= mix2_size/2)
  {
  mix2_tmp[2*i  ]=0;    
  mix2_tmp[2*i+1]=0;    
  mix2_tmp[2*(mix2_size-i)  ]=0;    
  mix2_tmp[2*(mix2_size-i)+1]=0;    
  i++;
  }
fftback(mix2_size, mix2_n, mix2_tmp, mix2_table, mix2_permute);
ia=mix2_size-cw_avg_points;
for(j=0; j<cw_avg_points; j++)
  {
  fftn_tmp[2*j  ]=mix2_tmp[ia];
  fftn_tmp[2*j+1]=0;
  ia+=2;
  }      
store_symmetry_adapted_dashdot(TRUE);
}

void coherent_cw_detect(void)
{
int i, j, ia;
float t1, t2, t3;
switch (cw_detect_flag)
  {
  case CWDETECT_CLEARED:
  if(cwspec_cnt < cwspec_max)return;
  cwspec_cnt=0;
// Initiate a new search for Morse code.
// We look for the "Morse code clock" in cw_spectrum.
  cwspec_pos[0]=evaluate_cw_spectrum();
  cwspec_ampl[0]=cw_spectrum[0];
  cw_detect_flag=CWDETECT_SEARCH_SPEED;
  cwspec_ptr=1;
  break;

  case CWDETECT_SEARCH_SPEED:
  if(cwspec_cnt < cwspec_max)return;
  cwspec_cnt=0;
// Collect more data from cw_spectrum.
  cwspec_pos[cwspec_ptr]=evaluate_cw_spectrum();
  cwspec_ampl[cwspec_ptr]=cw_spectrum[0];
  cwspec_ptr++;
// When CWSPEC_MAX spectra are evaluated, skip initial
// ones until the carrier power of the first one is above
// one third of the average carrier power.
  if(cwspec_ptr >= CWSPEC_MAX)
    {
    t1=cwspec_ampl[0];
    for(i=1; i<CWSPEC_MAX; i++)
      {
      t1+=cwspec_ampl[i];
      }
    t1/=3*CWSPEC_MAX;
    i=0;
    while(cwspec_ampl[i] < t1)i++;
    if(i != 0)
      {
      j=0;
      while(i<CWSPEC_MAX)
        {
        cwspec_ampl[j]=cwspec_ampl[i];
        cwspec_pos[j]=cwspec_pos[i];
        i++;
        j++;
        }
      cwspec_ptr=j;
      }
    else
      {
// We have CWSPEC_MAX peaks in cw_spectrum.
// Check if they are close together.
      t1=0;
      t2=0;
      for(i=0; i<CWSPEC_MAX; i++)
        {
        t1+=cwspec_ampl[i];
        t2+=cwspec_ampl[i]*cwspec_pos[i];
PRT00"\n%d  %f  %f  ",i,cwspec_ampl[i],cwspec_pos[i]);
        }
      t3=t2/t1;
      t1=0;
      t2=0;
      for(i=0; i<CWSPEC_MAX; i++)
        {
        if(fabs(t3-cwspec_pos[i]) < 0.1*cw_spectrum_size)
          {
          t1+=cwspec_ampl[i];
          t2+=cwspec_ampl[i]*cwspec_pos[i];
          }
        }
      if(t1 == 0)goto fail;
      t2/=t1;
      if(fabs(t3-t2) > 0.1*cw_spectrum_size)goto fail;
// The most probable frequency for the fundamental of the
// "Morse code clock" seems to be t2.
// Convert from frequency to time (samples)
      cwbit_pts=0.5*mix2_size/t2;
PRT00"\n cwbit_pts = %f (%f) ",cwbit_pts,cg_code_unit );
      cw_detect_flag=CWDETECT_COLLECT_RAMP;
      make_ideal_waveform();
      goto ramp;
      }
    }
  break;

  case CWDETECT_COLLECT_RAMP:
ramp:;

  collect_ramp();
DEB"\ncoherent: pc %d  [%d]    pe %d",baseb_pc,baseb_ramp[baseb_pc],baseb_pe);
  if( ((baseb_pc-baseb_pe+baseband_size)&baseband_mask) > 50*cwbit_pts)
    {
    cw_detect_flag=CWDETECT_SPEED;
    goto speed;
    }
  break;
    
  case CWDETECT_SPEED:
  collect_ramp();
speed:;
  detect_cw_speed();
  if(cw_detect_flag == CWDETECT_ERROR)goto fail;
  break;
        
  case CWDETECT_WAVEFORM_ESTABLISHED:
  collect_ramp();
  first_find_parts();
  break;

  case CWDETECT_SOME_PARTS_FITTED:
  collect_ramp();
  second_find_parts();
  break;

  case CWDETECT_LIMITS_FOUND:
  collect_ramp();
  init_cw_decode_region();
  break;

  case CWDETECT_REGION_INITIATED:
  collect_ramp();
  cw_decode_region();
  break;

  case CWDETECT_REGION_WAVEFORM_OK:
  collect_ramp();
  init_cw_decode();
  break;

  case CWDETECT_SOME_ASCII_FITTED:
  collect_ramp();
  cw_decode();
  break;

  case CWDETECT_DEBUG_STOP:
  lir_fillbox(screen_width/2,0,50,screen_height-1,3);
  DEB"\nno more!!");
  cw_detect_flag=CWDETECT_DEBUG_IDLE;
  break;

  case CWDETECT_DEBUG_IDLE:
  break;
  
  case CWDETECT_ERROR:
// Detect failed. Remove old data so buffers will not overflow.
fail:;
DEB"\nFAIL!!!!!!!!!!!!!!!!!!!!!");
  cw_detect_flag=CWDETECT_CLEARED;
// Do not store more than 2 minutes of bad data
// and make sure we leave some space in the buffer.  
PRT00"\n baseb_px= %d no_of_cwdat= %d",baseb_px, no_of_cwdat);  
  i=(baseb_pa-baseb_px+baseband_size)&baseband_mask;
  j=i;
  if(i>baseband_neg)i=baseband_neg;
  if(i>120*baseband_sampling_speed)i=120*baseband_sampling_speed;
  if(i != j)
    {
    ia=baseb_px;
    baseb_px=(baseb_pa-i+baseband_size)&baseband_mask;
DEB" new:  px= %d  no= %d",baseb_px, no_of_cwdat);  
    }
  no_of_cwdat=0;
  baseb_pf=baseb_px;
  baseb_pe=baseb_px;
  baseb_pd=baseb_px;
  baseb_pc=baseb_px;
  break;
  }
}

void check_cg_borders(void)
{
int xsiz, ysiz;
xsiz=3*cg_size+2;
ysiz=xsiz+5*text_height+4;
current_graph_minh=ysiz;
current_graph_minw=xsiz;
check_graph_placement((void*)(&cg));
set_graph_minwidth((void*)(&cg));
}


void help_on_coherent_graph(void)
{
int msg_no;
int event_no;
msg_no=68;
for(event_no=0; event_no<MAX_CGBUTT; event_no++)
  {
  if( cgbutt[event_no].x1 <= mouse_x &&
      cgbutt[event_no].x2 >= mouse_x &&
      cgbutt[event_no].y1 <= mouse_y &&
      cgbutt[event_no].y2 >= mouse_y)
    {
    switch (event_no)
      {
      case CG_TOP:
      case CG_BOTTOM:
      case CG_LEFT:
      case CG_RIGHT:
      msg_no=101;
      break;

      case CG_OSCILLOSCOPE: 
      msg_no=69;
      break;

      case CG_METER_GRAPH:
      msg_no=99;
      break;      
      }
    }
  }
help_message(msg_no);
}




void mouse_continue_coh_graph(void)
{
int i, j;
switch (mouse_active_flag-1)
  {
  case CG_TOP:
  if(cg.ytop!=mouse_y)goto cgm;
  break;

  case CG_BOTTOM:
  if(cg.ybottom!=mouse_y)goto cgm;
  break;

  case CG_LEFT:
  if(cg.xleft!=mouse_x)goto cgm;
  break;

  case CG_RIGHT:
  if(cg.xright==mouse_x)break;
cgm:;
  pause_screen_and_hide_mouse();
  graph_borders((void*)&cg,0);
  if(cg_oldx==COH_GRAPH_ACTIVE)
    {
    cg_oldx=mouse_x;
    cg_oldy=mouse_y;
    }
  else
    {
    i=mouse_x-cg_oldx;
    j=mouse_y-cg_oldy;
    cg_oldx=mouse_x;
    cg_oldy=mouse_y;
    cg.ytop+=j;
    cg.ybottom+=j;
    cg.xleft+=i;
    cg.xright+=i;
    check_cg_borders();
    }
  graph_borders((void*)&cg,15);
  resume_thread(THREAD_SCREEN);
  break;
    
  default:
  goto await_release;
  }
if(leftpressed == BUTTON_RELEASED)goto finish;
return;
await_release:;
if(leftpressed != BUTTON_RELEASED)return;
switch (mouse_active_flag-1)
  {
  case CG_OSCILLOSCOPE:
  pause_thread(THREAD_SCREEN);
  hide_mouse(0,screen_width>>1,0,screen_height);
  clear_cg_traces();
  resume_thread(THREAD_SCREEN);
  cg_oscill_on^=1;
  cg_oscill_on&=1;
  cg_max_trlevel=0;
  break;

  case CG_METER_GRAPH:
  pause_thread(THREAD_SCREEN);
  hide_mouse(0,screen_width>>1,0,screen_height);
  cg.meter_graph_on^=1;
  manage_meter_graph();
  resume_thread(THREAD_SCREEN);
  break;
  
  default:
  lirerr(872);
  break;
  }
finish:;
leftpressed=BUTTON_IDLE;
mouse_active_flag=0;
make_coherent_graph(TRUE);
cg_oldx=COH_GRAPH_ACTIVE;
}


void mouse_on_coh_graph(void)
{
int event_no;
// First find out is we are on a button or border line.
for(event_no=0; event_no<MAX_CGBUTT; event_no++)
  {
  if( cgbutt[event_no].x1 <= mouse_x &&
      cgbutt[event_no].x2 >= mouse_x &&
      cgbutt[event_no].y1 <= mouse_y &&
      cgbutt[event_no].y2 >= mouse_y)
    {
    cg_old_y1=cg.ytop;
    cg_old_y2=cg.ybottom;
    cg_old_x1=cg.xleft;
    cg_old_x2=cg.xright;
    mouse_active_flag=1+event_no;
    current_mouse_activity=mouse_continue_coh_graph;
    return;
    }
  }
// Not button or border.
// Do nothing.
current_mouse_activity=mouse_nothing;
mouse_active_flag=1;
}

void make_coherent_graph(int clear_old)
{
int len;
pause_thread(THREAD_SCREEN);
hide_mouse(0,screen_width>>1,0,screen_height);
clear_cg_traces();
if(clear_old)
  {
  hide_mouse(cg_old_x1,cg_old_x2,cg_old_y1,cg_old_y2);
  lir_fillbox(cg_old_x1,cg_old_y1,cg_old_x2-cg_old_x1+1,
                                                    cg_old_y2-cg_old_y1+1,0);
  }
check_cg_borders();
hide_mouse(cg.xleft,cg.xright,cg.ytop,cg.ybottom);  
scro[coherent_graph_scro].no=COH_GRAPH;
scro[coherent_graph_scro].x1=cg.xleft;
scro[coherent_graph_scro].x2=cg.xright;
scro[coherent_graph_scro].y1=cg.ytop;
scro[coherent_graph_scro].y2=cg.ybottom;
cgbutt[CG_LEFT].x1=cg.xleft;
cgbutt[CG_LEFT].x2=cg.xleft;
cgbutt[CG_LEFT].y1=cg.ytop;
cgbutt[CG_LEFT].y2=cg.ybottom;
cgbutt[CG_RIGHT].x1=cg.xright;
cgbutt[CG_RIGHT].x2=cg.xright;
cgbutt[CG_RIGHT].y1=cg.ytop;
cgbutt[CG_RIGHT].y2=cg.ybottom;
cgbutt[CG_TOP].x1=cg.xleft;
cgbutt[CG_TOP].x2=cg.xright;
cgbutt[CG_TOP].y1=cg.ytop;
cgbutt[CG_TOP].y2=cg.ytop;
cgbutt[CG_BOTTOM].x1=cg.xleft;
cgbutt[CG_BOTTOM].x2=cg.xright;
cgbutt[CG_BOTTOM].y1=cg.ybottom;
cgbutt[CG_BOTTOM].y2=cg.ybottom;
// Draw the border lines
graph_borders((void*)&cg,7);
make_button(cg.xleft+text_width-1,cg.ybottom-text_height/2-1,
                                               cgbutt,CG_OSCILLOSCOPE,'o');
make_button(cg.xleft+text_width-1,cg.ybottom-5*(text_height/2-1),
                                               cgbutt,CG_METER_GRAPH,'s');
cg_oldx=COH_GRAPH_ACTIVE;
settextcolor(7);
cg_x0=(cg.xleft+cg.xright)>>1;
len=((cg.xright-cg.xleft)>>1);
cg_y0=cg.ytop+len+1;
cg_y1=cg_y0+len;
lir_hline(cg.xleft+1,cg_y1,cg.xright-1,7);
if(kill_all_flag) return;
clear_coherent();
manage_meter_graph();
resume_thread(THREAD_SCREEN);
make_modepar_file(GRAPHTYPE_CG);
}

void init_coherent_graph(void)
{
int xsiz, ysiz;
zzpos=2*CG_AMAX+2;
xsiz=3*cg_size+2;
ysiz=xsiz+5*text_height+4;
if (read_modepar_file(GRAPHTYPE_CG) == 0)
  {
reinit:;
  cg.xright=screen_width-1;
  cg.xleft=cg.xright-xsiz;
  cg.ybottom=bg.ybottom;
  cg.ytop=cg.ybottom-ysiz;
  cg.meter_graph_on=1;
  }
if(cg.ybottom-cg.ytop != ysiz ||
   cg.xright-cg.xleft != xsiz)goto reinit;
cg_flag=1;
mg_flag=0;
cg_oscill_on=0;
if(genparm[AFC_ENABLE] !=0 && genparm[CW_DECODE_ENABLE] != 0)cg_oscill_on=1;
cg_osc_shift_flag=0;
cg_max_trlevel=0;
coherent_graph_scro=no_of_scro;
meter_graph_scro=no_of_scro+1;
if(no_of_scro+2 >= MAX_SCRO)lirerr(89);
make_coherent_graph(FALSE);
no_of_scro+=2;
}

int coh_tophat_filter(float *xin,float *xout,int lk,int lapt)
{
int i, k, len, sizhalf,first_point,last_point;
float t1,t2;
int ia,ib,ic;
len=lk|1;
sizhalf=mix2_size/2;
first_point=(lapt-sizhalf+baseband_size)&baseband_mask;
ia=(first_point+1-len+baseband_size)&baseband_mask;
ib=(first_point-len/2+baseband_size)&baseband_mask;
ic=first_point+1;
t1=0;
t2=0;
k=ia;
for(i=0; i<len; i++)
  {
  t1+=xin[2*k];
  t2+=xin[2*k+1];
  k=(k+1)&baseband_mask;
  }
last_point=(sizhalf+first_point)&baseband_mask;
while(ic!=last_point)
  {
  xout[2*ib  ]=t1/len;
  xout[2*ib+1]=t2/len;
  t1+=xin[2*ic  ];
  t2+=xin[2*ic+1];
  ic=(ic+1)&baseband_mask;
  t1-=xin[2*ia  ];
  t2-=xin[2*ia+1];
  ib=(ib+1)&baseband_mask;
  ia=(ia+1)&baseband_mask;
  }
xout[2*ib  ]=t1/len;
xout[2*ib+1]=t2/len;
return (ib+1)&baseband_mask;
}


