
#include <ctype.h>
#include "globdef.h"
#include "uidef.h"
#include "hwaredef.h"
#include "fft1def.h"
#include "screendef.h"
#include "seldef.h"
#include "sdrdef.h"
#include "thrdef.h"
#include "conf.h"

int hware_flag;
char hware_error_flag;
char hware_init_flag;
double hware_time;
int parport_status;
int parport_control;
int parport_ack;
int parport_ack_sign;
int sdr14_att_counter;
int sdr14_nco_counter;


#if (LUSERS_ROUTINES_PRESENT == 1 && OSNUM == OS_FLAG_LINUX)
#include <semaphore.h>
#include "lconf.h"
#include "ldef.h"
extern int serport;
#include "users_hwaredriver.c"
#else
#if (WUSERS_ROUTINES_PRESENT == 1 && OSNUM == OS_FLAG_WINDOWS)
#include "windef.h"
extern HANDLE serport;
#include "wusers_hwaredriver.c"
#else

void mouse_on_users_graph(void){}
void init_users_control_window(void){}
void users_init_mode(void){}
void users_eme(void){}
void userdefined_u(void){}
void userdefined_q(void){}


// **********************************************************
//                  WSE drive routines
// **********************************************************
// The radio hardware may be designed in many different ways.
// When the user selects a new frequency there may be several
// different hardware units that have to be changed.
// The routine supplied here is designed to work with the
// RXxxx -> RX70 -> RX10700 -> RX2500 converters.
// If you have some other hardware, you may replace or modify these 
// routines and move them into (w)users_hwaredriver.c   
//
// Associate a number to each frequency band that will
// need a separate case in set_hardware_frequency.
#define AUDIO_MODE 0
#define RX2500_TO_ANT 1
#define RX10700_TO_ANT 2
#define RX70_TO_ANT 3
#define RX144_TO_ANT 4
#define RXHFA_TO_ANT_1800 5
#define RXHFA_TO_ANT_3500 6
#define RXHFA_TO_ANT_7000 7
#define RXHFA_TO_ANT_10000 8
#define RXHFA_TO_ANT_14000 9
#define MAX_BANDS 10

#define TEST_GROUP_Y1 4
#define TEST_GROUP_Y2 TEST_GROUP_Y1 +10
#define TEST_GROUP_Y3 TEST_GROUP_Y2 +8

char *undefmsg={"Undef"};
char *device_name_rx10700={"RX10700"};
char *device_name_rx70={"RX70"};
char *device_name_rx144={"RX144"};
char *device_name_rxhfa={"RXHFA"};
char *device_name_rxhfa_gain={"Gain(RXHFA)"};

char device_bits[256];  
int device_data[256];
unsigned char hware_updates[256];
char *device_name[256];
int hware_retry;
int hware_device_no;
int hware_data;
int hware_bits;
int hware_no_of_updates;
char hware_outdat;


typedef struct {
double increment;
double lowest;
double highest;
int direction;
int label;
} FQ_INFO;

FQ_INFO fqinfo[MAX_BANDS]=
            {0.0  ,   0.0  ,   0.0  , 1, AUDIO_MODE,
             0.025,   1.675,   2.150,-1, RXHFA_TO_ANT_1800,
             0.025,   2.500,   2.500, 1, RX2500_TO_ANT,
             0.025,   3.475,   3.950,-1, RXHFA_TO_ANT_3500,
             0.025,   6.975,   7.450,-1, RXHFA_TO_ANT_7000,
             0.025,   9.975,  10.450,-1, RXHFA_TO_ANT_10000,
             0.025,  10.675,  10.750,-1, RX10700_TO_ANT,
             0.025,  13.975,  14.450,-1, RXHFA_TO_ANT_14000,
             0.025,  69.975,  70.450,-1, RX70_TO_ANT,
             0.025, 143.975, 145.950,-1, RX144_TO_ANT
             }; 
    




void users_close_devices(void)
{
// WSE units do not use the parallel port.
}

void users_open_devices(void)
{
if(ui.parport == 0)
  {
  allow_parport=FALSE;
  }
else
  {
  if(lir_parport_permission()==TRUE)
    {
    allow_parport=TRUE;
    }
  else  
    {
    allow_parport=FALSE;
    }
  }  
}

void hware_interface_test(void)
{
int i, pardat, j, k, xpos, bits[8];
int cntrldat;
char inpins[5]={15,13,12,10,11};
char s[80]; 
clear_screen();
clear_hware_data();
pardat=0;
cntrldat=0;
lir_text(3,TEST_GROUP_Y1,"DATA PORT (address, output only)");
lir_text(2,TEST_GROUP_Y1+2,"Enter 0 to 7 for bit to toggle");
lir_text(5,TEST_GROUP_Y1+4,"Bit:");
lir_text(2,TEST_GROUP_Y1+5,"Status:");
lir_text(5,TEST_GROUP_Y1+6,"Pin:");
lir_text(3,TEST_GROUP_Y3,"CONTROL PORT input pins");
lir_text(5,TEST_GROUP_Y3+3,"Bit:");
lir_text(2,TEST_GROUP_Y3+4,"Status:");
lir_text(5,TEST_GROUP_Y3+5,"Pin:");
lir_text(5,TEST_GROUP_Y2,"Enter D to toggle data (pin 17)");
lir_text(5,TEST_GROUP_Y2+1,"Enter C to toggle clock (pin 1)");
goto show;
loop:;
lir_refresh_screen();
lir_sleep(10000);
test_keyboard();
if(kill_all_flag) return;
if(lir_inkey != 0)
  {
  process_current_lir_inkey();
  if(kill_all_flag) return;
  if(lir_inkey >= '0' && lir_inkey <= '7')
    {
    k=1;
    j=lir_inkey-'0';
    while(j != 0)
      {
      k<<=1;
      j--;
      }
    pardat^=k;
    }
  if(lir_inkey=='D')
    {
    cntrldat ^= HWARE_DATA;
    }
  if(lir_inkey=='C')
    {
    cntrldat ^= HWARE_CLOCK;
    }
show:;  
  j=pardat;
  k=0;
  while(k<8)
    {
    bits[k]=j&1;
    if(bits[k] >0)bits[k]=1;
    j>>=1;
    k++;
    }
  xpos=11;
  for(k=0; k<8; k++)
    {
    sprintf(s,"%1d",k);
    lir_text(xpos,TEST_GROUP_Y1+4,s);
    sprintf(s,"%1d",bits[k]);
    lir_text(xpos,TEST_GROUP_Y1+5,s);
    sprintf(s,"%1d",k+2);
    lir_text(xpos,TEST_GROUP_Y1+6,s);
    xpos+=3;
    }
  for(i=0; i<20;i++)s[i]=' ';
  s[20]=0;  
  lir_text(xpos+5,TEST_GROUP_Y1+5,s);
  lir_text(xpos+5,TEST_GROUP_Y1+5,device_name[pardat]);
  s[1]=0;
  if( (cntrldat&HWARE_DATA) == 0)
    {
    s[0]='H';
    }
  else
    {
    s[0]='L';
    }  
  lir_text(3,TEST_GROUP_Y2+3,"Data=");
  lir_text(8,TEST_GROUP_Y2+3,s);
  if( (cntrldat&HWARE_CLOCK) == 0)
    {
    s[0]='H';
    }
  else
    {
    s[0]='L';
    }  
  lir_text(23,TEST_GROUP_Y2+3,"Clock=");
  lir_text(29,TEST_GROUP_Y2+3,s);
  lir_outb(pardat,ui.parport);
  lir_outb(cntrldat,parport_control);
  }
i=lir_inb(parport_status);
xpos=11;
i >>= 3;
j=8;
for(k=0; k<5; k++)
  {
  settextcolor(7);
  if( (j & parport_ack) !=0)settextcolor(12);
  sprintf(s,"%1d",k+3);
  lir_text(xpos+1,TEST_GROUP_Y3+3,s);
  sprintf(s,"%1d",i&1);
  lir_text(xpos+1,TEST_GROUP_Y3+4,s);
  sprintf(s,"%1d",inpins[k]);
  lir_text(xpos,TEST_GROUP_Y3+5,s);
  xpos+=3;
  i >>= 1;
  j <<= 1;
  }
settextcolor(7);
if(lir_inkey != 'X')goto loop;
}

void set_rxhfa_data(int band)
{
if(device_data[RX144] != band)
  {
  device_data[RXHFA]=band;
  device_bits[RXHFA]=5;
  hware_updates[hware_no_of_updates]=RXHFA;
  hware_no_of_updates++;
  }
}

int set_rx144_data(int stepno)
{
int i;
i=8;
while(stepno >= 20)
  {
  i>>=1;
  stepno-=20;
  }
if(device_data[RX144] != i)
  {
  device_data[RX144]=i;
  device_bits[RX144]=4;
  hware_updates[hware_no_of_updates]=RX144;
  hware_no_of_updates++;
  }
return stepno;  
}

void set_rx10700_data(int stepno)
{
int i;
i=8;
while(stepno != 0)
  {
  i>>=1;
  stepno--;
  }
if(device_data[RX10700] != i)
  {
  device_data[RX10700]=i;
  device_bits[RX10700]=4;
  hware_updates[hware_no_of_updates]=RX10700;
  hware_no_of_updates++;
  }
}

int set_rx70_data(int stepno)
{
int i;
i=16;
while(stepno >= 4)
  {
  i>>=1;
  stepno-=4;
  }
if(device_data[RX70] != i)
  {
  device_data[RX70]=i;
  device_bits[RX70]=5;
  hware_updates[hware_no_of_updates]=RX70;
  hware_no_of_updates++;
  }
return stepno;
}

void set_hardware_rx_gain(int band_no)
{
// bit0 to bit 2 for amplifier/bypass/-10 dB.
// bit0-bit2 = 6 => -10dB
// bit0-bit2 = 3 => 0dB
// bit0-bit2 = 5 => +10dB
//
// bit3 for 5 dB attenuator.
// bit3 = 0 => -5dB
// bit3 = 1 =>  0dB
 
int gaindat[6]={6,14,3,11,5,13};
int i;
if(ui.ad_device_no == SDR14_DEVICE_CODE)
  {
  fg.gain_increment=10;
  fg.gain/=10;
  fg.gain*=10;
  if(fg.gain > 0)fg.gain=0;
  if(fg.gain<-30)fg.gain=-30;
  if(hware_init_flag==0)sdr14_att_counter++;
  return;
  }
if( ui.parport == 0 || !allow_parport)return;
switch (fqinfo[band_no].label)
  {   
  case RXHFA_TO_ANT_1800:
  case RXHFA_TO_ANT_3500:
  case RXHFA_TO_ANT_7000:
  case RXHFA_TO_ANT_10000:
  case RXHFA_TO_ANT_14000:
  fg.gain_increment=5;
  fg.gain/=5;
  fg.gain*=5;
  if(fg.gain < -15)fg.gain=-15;
  if(fg.gain > 10)fg.gain=10;
  i=fg.gain/5+3;
  if(device_data[RXHFA_GAIN] != gaindat[i])
    {
    device_data[RXHFA_GAIN]=gaindat[i];
    device_bits[RXHFA_GAIN]=4;
    if(hware_no_of_updates<0)hware_no_of_updates=0;
    hware_updates[hware_no_of_updates]=RXHFA_GAIN;
    hware_no_of_updates++;
    hware_flag=1;
    hware_time=current_time();
    }
  break;
  
  default:
  break;
  }
}

void set_hardware_rx_frequency(void)
{
float t1;
int band_no, stepno;
if(ui.ad_device_no == SDR14_DEVICE_CODE)
  {
  if(hware_init_flag==0)sdr14_nco_counter++;
  fft1_direction=fg.passband_direction=1;
  fg.passband_increment=.025;
  return;
  }
if( ui.parport == 0 || !allow_parport)return;
// First of all, stop any ongoing activities on the parallel port.
lir_outb(0,ui.parport);
lir_outb(HWARE_CLOCK,parport_control);
// Flag unwritten data as not valid
while(hware_no_of_updates >= 0)
  {
  device_data[hware_updates[hware_no_of_updates]]=-1;
  hware_no_of_updates--;
  }
hware_no_of_updates=0;
hware_flag=0;
band_no=0;
while(band_no < MAX_BANDS-1 && 
     fg.passband_center > fqinfo[band_no].highest)band_no++;
if(fg.passband_center < fqinfo[band_no].lowest && band_no > 0) 
  {
  t1=0.5*(fqinfo[band_no].lowest+fqinfo[band_no-1].highest);
  if(fg.passband_center < t1)band_no--;
  }
if(fg_old_band != band_no)set_hardware_rx_gain(band_no);
fg_old_band=band_no;
t1=(fg.passband_center-fqinfo[band_no].lowest)/fqinfo[band_no].increment;
stepno=t1+0.5;
if(stepno < 0)stepno=0;
set_fq:;
fg.passband_center=fqinfo[band_no].lowest+stepno*fqinfo[band_no].increment;  
if(fg.passband_center > fqinfo[band_no].highest+0.0000001)
  {
  stepno--;
  goto set_fq;
  }  
fg.passband_increment = fqinfo[band_no].increment;
fg.passband_direction = fqinfo[band_no].direction;
fft1_direction=fg.passband_direction;
// Send a message to the screen thread to redraw the frequency scale
// on the main spectrum.
sc[SC_WG_FQ_SCALE]++;
// Set up data to switch relays to connect units properly.
// and to select the correct LO frequency for all units in use.  
switch (fqinfo[band_no].label)
  {   
  case AUDIO_MODE:
// Send nothing to the hardware, just set frequency
// and direction to zero so frequency scale in waterfall and
// other graphs gives audio frequency.
  break;
     
  case RX2500_TO_ANT:
// Route the signal source to the RX2500 input. 
  break;
  
  case RX10700_TO_ANT:
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RX10700 input. 
  goto set_10700;
  
  case RX70_TO_ANT:
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RX10700 input. 
  goto set_70;

  case RXHFA_TO_ANT_1800:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(16);
  goto set_70;

  case RXHFA_TO_ANT_3500:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(8);
  goto set_70;

  case RXHFA_TO_ANT_7000:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(4);
  goto set_70;

  case RXHFA_TO_ANT_10000:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(2);
  goto set_70;

  case RXHFA_TO_ANT_14000:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(1);
  goto set_70;

  case RX144_TO_ANT:
// Route RX144 output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RX10700 input. 
  stepno=set_rx144_data(stepno);
set_70:;
  stepno=set_rx70_data(stepno);
set_10700:;
  set_rx10700_data(stepno);
  break;
  }
if(hware_no_of_updates != 0)
  {
  hware_flag=1;
  hware_time=current_time();
  }
else
  {
  hware_init_flag=0;
  }
}

void clear_hware_data(void)
{
int i;
// Set -1 in device_data. 
// A data word is max 31 bit, the sign bit is a flag
// saying the data is not valid.
for(i=0; i<256; i++)
  {
  device_data[i]=-1;
  device_name[i]=undefmsg;
  }
device_name[RX10700]=device_name_rx10700;
device_name[RX70]=device_name_rx70;
device_name[RX144]=device_name_rx144;
device_name[RXHFA]=device_name_rxhfa;
device_name[RXHFA_GAIN]=device_name_rxhfa_gain;
hware_init_flag=1;
hware_no_of_updates=0;
hware_flag=0;
}

void control_hware(void)
{
int err;
unsigned char outdat;
char s[80];
switch(hware_flag)
  {
  case 1:
  hware_no_of_updates--;
  if(hware_no_of_updates <0)
    {
    lirerr(988342);
    return;
    }
  hware_device_no=hware_updates[hware_no_of_updates];
  hware_error_flag=0;
  hware_data=device_data[hware_device_no];
  hware_bits=device_bits[hware_device_no];
// A new unit is selected.
// Write the number of this unit to the data port
  outdat=hware_device_no;
  lir_outb(outdat,ui.parport);
  hware_flag=2;
// Set clock low and make data the first bit of current output.
  if( (hware_data & 1) == 0)
    {
    hware_outdat=HWARE_DATA+HWARE_CLOCK;
    }
  else
    {
    hware_outdat=HWARE_CLOCK;
    }  
  break;

  case 2:
// Set clock high to clock in new data bit.
  hware_outdat^=HWARE_CLOCK;
  hware_flag=3;
  hware_retry=0;
  break;

  case 3:
// Read status to see if we really got the data bit into the unit.
  outdat=lir_inb(parport_status)&parport_ack;
  outdat^=parport_ack_sign;
  err=0;
  if(outdat == 0)
    {
    if( (hware_outdat&HWARE_DATA) != 0)
      {
      err=1;
      }
    }
  else  
    {
    if( (hware_outdat&HWARE_DATA) == 0)
      {
      err=1;
      }
    }
  hware_retry++;
  if(err==1)
    {
    if(hware_retry<4)return;
    hware_error_flag=1;
    }
  hware_bits--;
  if(hware_bits == 0)
    {
// No more. Deselect unit.
    lir_outb(0,ui.parport);
    hware_flag=4;
    return;
    }
  else
    {
// Set clock low and make data the next bit of current output.
    hware_data=hware_data>>1;
    if( (hware_data & 1) == 0)
      {
      hware_outdat=HWARE_DATA+HWARE_CLOCK;
      }
    else
      {
      hware_outdat=HWARE_CLOCK;
      }  
    hware_flag=2;
    }
  break;

  case 4:
// Select unit. This time we actually clock the new data 
// from the shift register into the latch.
  outdat=hware_device_no;
  lir_outb(outdat,ui.parport);
  hware_flag=5;
  return;

  case 5:
  lir_outb(0,ui.parport);
  if(hware_no_of_updates > 0)
    {
    hware_flag=1;
    }
  else
    {  
    hware_flag=0;
    hware_init_flag=0;
    hware_no_of_updates=-1;
    }
  hware_outdat=HWARE_CLOCK;
  if(hware_error_flag != 0)
    {
    sprintf(s,"ERROR: %s",device_name[hware_device_no]);
    wg_error(s,5);
    }

  break;      
  }
lir_outb(hware_outdat,parport_control);
}
// ********************************************************
//               End WSE routines
// ******************************************************** 
#endif
#endif
