// users_hwaredriver.c for external frequency control via serial port
// 31 march 2007
// written for linrad02-27
// to select a transceiver uncomment the corresponding define statement
// runs under windows and linux 
// To generate the windows version from linux:
//   install mingw32 from Leif's website,
//   save this users_hwaredriver.c as wusers_hwaredriver.c 
//   run ./configure and  make linrad.exe
//
// based on work from dl3iae/ik7ezn/on4iy/on7eh/k1jt/w3sz
// All errors are mine
// Use at your own risk and enjoy
// Pierre/ON5GN


//#define TRANSCEIVER_NAME "IC-275"
#define TRANSCEIVER_NAME "IC-706"

//#define TRANSCEIVER_NAME "FT-1000"
//#define TRANSCEIVER_NAME "FT-736"

//#define TRANSCEIVER_NAME "TS-850"
//#define TRANSCEIVER_NAME "TS-2000"


// ******************************************************************************
//               ROUTINES FOR A 'MOVABLE' FIXED SIZE USERGRAPH
//             screenposition of usergraph and offset values are 
//                saved in the file par_"TRANSCEIVER_NAME"
//                 The same file is used in all rx modes. 
// ******************************************************************************


#include <string.h>

#define USERS_GRAPH_TYPE1 64
typedef struct {
int ytop;
int ybottom;
int xleft;
int xright;
int par1;
int par2;
int yborder;    // required for move graph
} UG_PARMS;

#define UG_TOP 0
#define UG_BOTTOM 1
#define UG_LEFT 2
#define UG_RIGHT 3
#define UG_INCREASE_PAR1 4
#define UG_DECREASE_PAR1 5
#define MAX_UGBUTT 6

UG_PARMS ug;
BUTTONS ugbutt[MAX_UGBUTT];

int ug_old_y1;
int ug_old_y2;
int ug_old_x1;
int ug_old_x2;
int users_graph_scro;

int ug_oldx;     // required for move graph
int ug_oldy;     // required for move graph
int ug_yborder;  // required for move graph
int ug_hsiz;     // required for move graph
int ug_vsiz;     // required for move graph 

int ug_xleft_init ;
int ug_ytop_init;

char ug_msg0[80]; //messages in usergraph 
char ug_msg1[80];
char ug_msg2[80];
int ug_msg_red;   //switch for colormanagement of messages in usergraph
char ugfile[20];  //name of usergraph parameter file

void check_ug_borders(void)  // required for move graph
{
current_graph_minh=ug_vsiz;
current_graph_minw=ug_hsiz;
check_graph_placement((void*)(&ug));
set_graph_minwidth((void*)(&ug));
}

void show_user_parms(void)
{
char s[80];
FILE *Fp;
// Show the user parameters on screen
// and issue hardware commands if required.
// Use hware_flag to direct calls to control_hware() when cpu time 
// is available in case whatever you want to do takes too much time 
// to be done immediately.
// Note that mouse control is done in the narrowband processing thread
// and that you can not use lir_sleep to wait for hardware response here.
hide_mouse(ug.xleft, ug.xright,ug.ytop,ug.ybottom);

sprintf(s," %s EXT. FREQ. CONTROL ",TRANSCEIVER_NAME);
lir_pixwrite(ug.xleft+2*text_width,ug.ytop+.4*text_height,s);

sprintf(s,"Offset-fine(Hz): %d ",ug.par1);
lir_pixwrite(ug.xleft+4*text_width,ug.ytop+1.5*text_height,s);
sprintf(s,"Offset-raw(Khz): %d ",ug.par2);
lir_pixwrite(ug.xleft+4*text_width,ug.ytop+1+2.5*text_height,s);

//DISPLAY MESSAGES IN  MESSAGE-AREA OF USERGRAPH

//truncate message if length  > 29 characters
if (strlen(ug_msg0) > 29) ug_msg0[29]='\0';
if (strlen(ug_msg1) > 29) ug_msg1[29]='\0';
if (strlen(ug_msg2) > 29) ug_msg2[29]='\0';

if (ug_msg_red==1)
{settextcolor(12);
}
else
{
settextcolor(10);
}
lir_pixwrite(ug.xleft+31*text_width,ug.ytop+.5*text_height,ug_msg0);
lir_pixwrite(ug.xleft+31*text_width,ug.ytop+1.5*text_height,ug_msg1);
lir_pixwrite(ug.xleft+31*text_width,ug.ytop+2.5*text_height,ug_msg2);
settextcolor(7);

//save screenposition and offset-values in usergraph to usergraph parameter file
Fp=fopen(ugfile,"w");
fprintf(Fp,"%d\n%d\n%d\n%d\n",ug.xleft,ug.ytop,ug.par1,ug.par2);
fclose(Fp);
}

void make_users_control_graph(void)
{
pause_thread(THREAD_SCREEN);
// Set a negative number here.
// In case several windows are desired, give them different
// negative numbers and use scro[m].no to decide what to do
// in your mouse_on_users_graph routine.
check_ug_borders();  // required for move graph
scro[users_graph_scro].no=USERS_GRAPH_TYPE1;
// These are the coordinates of the border lines.
scro[users_graph_scro].x1=ug.xleft;
scro[users_graph_scro].x2=ug.xright;
scro[users_graph_scro].y1=ug.ytop;
scro[users_graph_scro].y2=ug.ybottom;
// Each border line is treated as a button.
// That is for the mouse to get hold of them so the window can be moved.
ugbutt[UG_LEFT].x1=ug.xleft;
ugbutt[UG_LEFT].x2=ug.xleft;
ugbutt[UG_LEFT].y1=ug.ytop;
ugbutt[UG_LEFT].y2=ug.ybottom;
ugbutt[UG_RIGHT].x1=ug.xright;
ugbutt[UG_RIGHT].x2=ug.xright;
ugbutt[UG_RIGHT].y1=ug.ytop;
ugbutt[UG_RIGHT].y2=ug.ybottom;
ugbutt[UG_TOP].x1=ug.xleft;
ugbutt[UG_TOP].x2=ug.xright;
ugbutt[UG_TOP].y1=ug.ytop;
ugbutt[UG_TOP].y2=ug.ytop;
ugbutt[UG_BOTTOM].x1=ug.xleft;
ugbutt[UG_BOTTOM].x2=ug.xright;
ugbutt[UG_BOTTOM].y1=ug.ybottom;
ugbutt[UG_BOTTOM].y2=ug.ybottom;
// Draw the border lines
graph_borders((void*)&ug,7);
ug_oldx=-10000;                   //from freq_control
settextcolor(7);
make_button(ug.xleft+27.5*text_width,ug.ybottom-3*text_height/2-2,
                                         ugbutt,UG_DECREASE_PAR1,25);
make_button(ug.xleft+29*text_width,ug.ybottom-3*text_height/2-2,
                                     ugbutt,UG_INCREASE_PAR1,24); 
//make_button(ug.xleft+29*text_width,ug.ytop+text_height-2,
//                                         ugbutt,UG_DO_WHATEVER,'Q');
show_user_parms();
//draw separatorlines in usergraph
lir_line(ug.xleft+30*text_width,ug.ytop,ug.xleft+30*text_width,ug.ytop+3.5*text_height,7);
lir_line(ug.xleft,ug.ytop+1.3*text_height,ug.xleft+30*text_width,ug.ytop+1.3*text_height,7);
//
resume_thread(THREAD_SCREEN);
}


void mouse_continue_users_graph(void)
{
int i, j;      // required for move graph

switch (mouse_active_flag-1)
  {
// Move border lines immediately.
// for other functions, wait until button is released.
// Move fixed size window  based on coding in freq_control.c
case UG_TOP:
  if(ug.ytop!=mouse_y)goto ugm;
  break;

  case UG_BOTTOM:
  if(ug.ybottom!=mouse_y)goto ugm;
  break;

  case UG_LEFT:
  if(ug.xleft!=mouse_x)goto ugm;
  break;

  case UG_RIGHT:
  if(ug.xright==mouse_x)break;
ugm:;
  pause_screen_and_hide_mouse();
  graph_borders((void*)&ug,0);
  if(ug_oldx==-10000)
    {
    ug_oldx=mouse_x;
    ug_oldy=mouse_y;
    }
  else
    {
    i=mouse_x-ug_oldx;
    j=mouse_y-ug_oldy;  
    ug_oldx=mouse_x;
    ug_oldy=mouse_y;
    ug.ytop+=j;
    ug.ybottom+=j;
    ug.xleft+=i;
    ug.xright+=i;
    check_ug_borders();    // check borders
    ug.yborder=(ug.ytop+ug.ybottom)>>1;
    }
  graph_borders((void*)&ug,15); 
  resume_thread(THREAD_SCREEN);
  break;


  default:
  goto await_release;
  }
if(leftpressed == BUTTON_RELEASED)goto finish;
return;
await_release:;
if(leftpressed != BUTTON_RELEASED) return;
// Assuming the user wants to control hardware we
// allow commands only when data is not over the network.
//if(network_flag != 2)        //for linrad02.20 
if((ui.network_flag&NET_RX_INPUT) == 0) //for linrad02.23
  {
  switch (mouse_active_flag-1)
    {
    case UG_INCREASE_PAR1: 
    ug.par1++;
    break;
  
    case UG_DECREASE_PAR1:
    ug.par1--;
    break;

    default:
// This should never happen.    
    lirerr(211053);
    break;
    }
  }  
finish:;
hide_mouse(ug_old_x1,ug_old_x2,ug_old_y1,ug_old_y2);
leftpressed=BUTTON_IDLE;  
mouse_active_flag=0;
graph_borders((void*)&ug,0);
lir_fillbox(ug_old_x1,ug_old_y1,ug_old_x2-ug_old_x1,ug_old_y2-ug_old_y1,0);
make_users_control_graph();
ug_oldx=-10000; 
}

void new_user_par1(void)
{
ug.par1=numinput_int_data;
pause_thread(THREAD_SCREEN);
show_user_parms();
resume_thread(THREAD_SCREEN);
}

void new_user_par2(void)
{
ug.par2=numinput_float_data;
pause_thread(THREAD_SCREEN);
show_user_parms();
resume_thread(THREAD_SCREEN);
}


void mouse_on_users_graph(void)
{
int event_no;
// First find out if we are on a button or border line.
for(event_no=0; event_no<MAX_UGBUTT; event_no++)
  {
  if( ugbutt[event_no].x1 <= mouse_x && 
      ugbutt[event_no].x2 >= mouse_x &&      
      ugbutt[event_no].y1 <= mouse_y && 
      ugbutt[event_no].y2 >= mouse_y) 
    {
    ug_old_y1=ug.ytop;
    ug_old_y2=ug.ybottom;
    ug_old_x1=ug.xleft;
    ug_old_x2=ug.xright;
    mouse_active_flag=1+event_no;
    current_mouse_activity=mouse_continue_users_graph;
    return;
    }
  }
// Not button or border.
// Prompt the user for par1 or par2 depending on whether the mouse is
// in the upper or the lower part of the window.
mouse_active_flag=1;
numinput_xpix=ug.xleft+21*text_width;
if(mouse_x > ug.xleft+20*text_width && mouse_x<ug.xright-32*text_width)
  {
  if(mouse_y > (ug.ytop+2*text_height))
    {
    numinput_ypix=ug.ytop+(2.5*text_height)+1;
    numinput_chars=7;
    erase_numinput_txt();
    numinput_flag=FIXED_FLOAT_PARM;
    par_from_keyboard_routine=new_user_par2; //offset-raw
    }
  else
    {
    numinput_ypix=ug.ytop+1.5*text_height;
    numinput_chars=4;
    erase_numinput_txt();
    numinput_flag=FIXED_INT_PARM;
    par_from_keyboard_routine=new_user_par1; //offset-fine
    }  
  }
else
  {
// If we did not select a numeric input by setting numinput_flag
// we have to set mouse_active flag.
// Set the routine to mouse_nothing, we just want to
// set flags when the mouse button is released.    
  current_mouse_activity=mouse_nothing;
  mouse_active_flag=1;
  }
}

void init_users_control_window(void)
{
// Set graph position and offset values by  reading usergraph  file
// 
FILE *Fp;
sprintf(ugfile,"par_%s_ug",TRANSCEIVER_NAME);
Fp=fopen(ugfile,"r");
if (Fp==NULL)          // if file not found, use default-values 
{

ug_xleft_init =60*text_width;                     //default-value for initial x position 
ug_ytop_init = (screen_last_line-4)*text_height;  //default-value for initial y position

ug.par1=0;                                        //default-value for initial offset-fine
ug.par2=0;                                        //default-value for initial offset-raw
}
else
{
fscanf(Fp,"%d%d%d%d",&ug_xleft_init,&ug_ytop_init,&ug.par1,&ug.par2);
fclose(Fp);
}
ug_msg_red=1;
strcpy(ug_msg0," HIT Q-KEY TO SET FREQUENCY  ");   // STARTUP MESSAGE
strcpy(ug_msg1,"                             ");   //CLEAR MESSAGE-AREA OF USER GRAPH
strcpy(ug_msg2,"                             ");   //CLEAR MESSAGE-AREA OF USER GRAPH
ug_hsiz=(60*text_width);                           // required for move graph
ug_vsiz=(3.5*text_height);                         // required for move graph
ug.xleft=ug_xleft_init;
ug.xright=ug_xleft_init+ug_hsiz;
ug.ytop= ug_ytop_init;
ug.ybottom= ug_ytop_init+ug_vsiz;
users_graph_scro=no_of_scro;
make_users_control_graph();
no_of_scro++;
if(no_of_scro >= MAX_SCRO)lirerr(89);
}



void users_init_mode(void)
{
// A switch statement can be used to do different things depending on
// the processing mode.
//switch (rx_mode)
//  {
//  case MODE_WCW:
//  case MODE_HSMS:
//  case MODE_QRSS:
//  break;
//  case MODE_NCW:
//  case MODE_SSB:
// Open a window to allow mouse control of your own things
  init_users_control_window();  
//  break;
//  case MODE_FM:
//  case MODE_AM:
//  case MODE_TXTEST:
//  case MODE_RX_ADTEST:
//  case MODE_TUNE:
//  break;
//  }
}
// *************************************************************
//          END ROUTINES FOR 'MOVABLE' FIXED SIZE USERGRAPH
// *************************************************************

void users_eme(void)
{
}
/*
void hware_interface_test(void)
{
}
*/
void userdefined_u(void)
{
}

void userdefined_q(void)
{
// **************************************************************************
//    ROUTINES FOR EXTERNAL FREQUENCY CONTROL VIA SERIAL PORT 
//     These routines are  called when the 'Q' key is pressed
// **************************************************************************

extern char serport_name[10];
int n=-1;
int i;
char s[80];
double transceiver_freq;
double min_transceiver_freq=0;
double max_transceiver_freq=0;
double offset_hwfreq;
double transverter_offset_raw; //raw offset in Khz (for tranverter) 
double transverter_offset_fine; //fine offset in Hz (for transverter)
char transceiver_name[10];
int serport_baudrate=0;
int serport_stopbits=0;
unsigned char a;
unsigned long freq;

//INITIALIZE

//initialise message-area
ug_msg_red=1;
strcpy(ug_msg0,"FREQUENCY SETTING IN PROGRESS");
strcpy(ug_msg1,"                             ");
strcpy(ug_msg2,"                             ");
show_user_parms();

//initialise transceiver_name
sprintf (transceiver_name,"%s ",TRANSCEIVER_NAME);

//initialize serial port parameters and frequency-range
if (strncmp(transceiver_name,"IC-275",6)==0)
{
serport_baudrate=9600;
serport_stopbits=0;
min_transceiver_freq=144000000;
max_transceiver_freq=146000000;
}

if (strncmp(transceiver_name,"IC-706",6)==0)
{
serport_baudrate=9600;
serport_stopbits=0;
min_transceiver_freq=1600000;
max_transceiver_freq=146000000;
}

if (strncmp(transceiver_name,"TS-850",6)==0)
{
serport_baudrate=4800;
serport_stopbits=1;
min_transceiver_freq=1600000;
max_transceiver_freq=30000000;
}

if (strncmp(transceiver_name,"TS-2000",7)==0)
{
serport_baudrate=4800;
serport_stopbits=1;
min_transceiver_freq=1600000;
max_transceiver_freq=148000000;
}

if (strncmp(transceiver_name,"FT-736",6)==0)
{
serport_baudrate=4800;
serport_stopbits=1;
min_transceiver_freq=144000000;
max_transceiver_freq=146000000;
}

if (strncmp(transceiver_name,"FT-1000",7)==0)
{
serport_baudrate=4800;
serport_stopbits=1;
min_transceiver_freq=1600000;
max_transceiver_freq=30000000;
}

//EXIT IF NO LINRAD-FREQUENCY IS SELECTED
 
if(mix1_selfreq[0] < 0)  
{
  ug_msg_red=1;
  sprintf (ug_msg0,"NO FREQ. SELECT. ON W/F GRAPH");
  goto userdefined_q_exit;
}

// EXIT IF INVALID  TRANSCEIVER FREQUENCY 

transverter_offset_raw=ug.par2;          //raw offset in Khz for transverter 
transverter_offset_fine=ug.par1;         //fine offset in Hz for transverter 
offset_hwfreq= transverter_offset_raw +(transverter_offset_fine*.001) ; //(Frequencies in KiloHertz !)
transceiver_freq= (hwfreq+offset_hwfreq)*1000;
transceiver_freq=floor(transceiver_freq);                               //Frequency in Hertz !

if ((transceiver_freq < min_transceiver_freq )
     ||(transceiver_freq > max_transceiver_freq))
{
  ug_msg_red=1;
  sprintf (ug_msg0,"INVALID %s FREQ.                 ",TRANSCEIVER_NAME); 
  sprintf (ug_msg1,"TOTAL OFFSET: %10.0f Hz",offset_hwfreq*1000); 
  sprintf (ug_msg2,"%s FREQ.:%10.0f Hz",TRANSCEIVER_NAME,transceiver_freq);
  goto userdefined_q_exit;
}

//OPEN SERIAL PORT 

lir_open_serport(1,                     /* 1=COM1 or /dev/ttyS0 , 2=COM2.... */
                 serport_baudrate,      /* baudrate 110 to 57600 */
                 serport_stopbits       /* 1=two stopbits, 0=one stopbit*/
                 );
ug_msg_red=0;
sprintf (ug_msg0,"OPEN %s PORT OK                      ",serport_name );


// SEND 'SET FREQUENCY COMMAND' TO CAT SYSTEM OF TRANSCEIVER


//***********************IC-275/IC-706 ROUTINES******************** 

if ((strncmp(transceiver_name,"IC-275",6)==0)
    |(strncmp(transceiver_name,"IC-706",6)==0))
{

struct {            /*'Set Frequency message' for IC-275/IC-706 */
   char prefix[2];
   char radio_addr;
   char computer_addr;
   char cmd_nr;
   char freq[5];
   char suffix;
   } msg0 = {0xFE, 0xFE,               /* cmd mode*/
             0x10,                     /* target address for IC-275*/
             0xE0,                     /* your computer address:customise to your environment*/
             0x05,                     /* change freq cmd */
             0x00,0x00,0x00,0x00,0x00, /* freq: x00 x89 x67 x45 x01 => 145.678.900 hz */
             0xFD};                    /* end cmd */

if (strncmp(transceiver_name,"IC-706",6)==0) 
sprintf(&msg0.radio_addr,"%c",0x58);   /* target address for IC-706 */

//convert 10 digits of transceiver_freq to packed bcd into msg0.freq
freq= transceiver_freq;
for (i=0; i < 5; i++) {
   a = freq%10;
   freq /= 10;
   a |= (freq%10)<<4;
   freq /= 10;
   msg0.freq[i]= a;
//trace output to screen
//printf("\n %i ",msg0.freq[i]);
   }

// write 'Set Frequency message' to serial port
n=lir_write_serport((char *) &msg0, sizeof(msg0));
}

//***********************TS850/TS-2000 ROUTINE****************


if ((strncmp(transceiver_name,"TS-850",6)==0)
    |(strncmp(transceiver_name,"TS-2000",7)==0))
{
// "VFO-A frequency" command 
// Freq in Hz; 11 digits, leading zeros, no decimal point, trailing semicolon.
sprintf (s,"FA%011.0f;",transceiver_freq);
n=lir_write_serport(s,14);
}

//************************FT736 ROUTINES**********************

if (strncmp(transceiver_name,"FT-736",6)==0)
{
unsigned char open_cmnd[5] =  { 0x00, 0x00, 0x00, 0x00, 0x00};
unsigned char close_cmnd[5] = { 0x80, 0x80, 0x80, 0x80, 0x80};
unsigned char setfreq_cmnd[5]= { 0x00, 0x00, 0x00, 0x00, 0x01}; /*x14 x50 x00 x00 => 145.000.000 HZ */
unsigned char cat_cmnd[5];

int write_cat_cmnd(void)
{
int r;
for (i=0;i<5;i=i+1)
 {
  r=lir_write_serport((char*)&cat_cmnd[i],1);
  if (r<0) return -1;
//trace output to screen
// printf("\n %i ",cat_cmnd[i]);
  lir_sleep(50000); // wait 50msec between each byte
 }
return 0;
}

//open CAT system
memcpy(cat_cmnd,open_cmnd, 5);
n=write_cat_cmnd();
if (n<0) goto ft736_exit;

//convert 8 digits of transceiver_freq to packed bcd into setfreq_cmd
freq= transceiver_freq;
freq /= 10;
for (i=0; i < 4; i++) {
   a = freq%10;
   freq /= 10;
   a |= (freq%10)<<4;
   freq /= 10;
   setfreq_cmnd[3-i]= a;
//trace output to screen
//printf("\n %i ",setfreq_cmnd[3-i]);
   }

//write setfreq_cmd to the serial port
memcpy(cat_cmnd,setfreq_cmnd, 5);
n=write_cat_cmnd();
if (n<0) goto ft736_exit;

//close CAT system
memcpy(cat_cmnd,close_cmnd, 5);
n=write_cat_cmnd();
ft736_exit:;
}

//***************************FT-1000 ROUTINES*********************

if (strncmp(transceiver_name,"FT-1000",7)==0)
{
unsigned char setfreq_cmnd[5]= { 0x00, 0x00, 0x00, 0x00, 0x0a}; /*x00 x50 x42 x01 => 142.500.000 hz */

//convert 8 digits of transceiver_freq to packed bcd into setfreq_cmd
freq= transceiver_freq;
freq /= 100;
for (i=0; i < 4; i++) {
   a = freq%10;
   freq /= 10;
   a |= (freq%10)<<4;
   freq /= 10;
   setfreq_cmnd[i]= a;
//trace output to screen
//printf("\n %i ",setfreq_cmnd[i]);
   }

//write setfreq_cmnd to the serial port
for (i=0;i<5;i=i+1)
 {
  n=lir_write_serport((char*)&setfreq_cmnd[i],1);
  if (n<0)  goto ft1000_exit;
//trace output to screen
// printf("\n %i ",setfreq_cmnd[i]);
  lir_sleep(5000); // wait 5msec between each write
 }
ft1000_exit:;
}

//DISPLAY RESULT IN USERGRAPH 

if (n<0)
  {
  ug_msg_red=1;
  sprintf (ug_msg0,"WRITE TO %s FAILED                  ",serport_name);
  lir_close_serport();
  goto userdefined_q_exit;
  }

ug_msg_red=0;
sprintf (ug_msg0,"WRITE TO %s PORT OK                    ",serport_name);
sprintf (ug_msg1,"TOTAL OFFSET: %10.0f Hz",offset_hwfreq*1000); 
sprintf (ug_msg2,"%s FREQ.:%10.0f Hz",TRANSCEIVER_NAME,transceiver_freq);
lir_close_serport();

userdefined_q_exit:;
show_user_parms();
}
// **************************************************************************
// **** END ROUTINES FOR FREQUENCY SETTING OF TRANSCEIVER VIA SERIAL PORT ***
// **************************************************************************
/*
void users_close_devices(void)
{
}

void users_open_devices(void)
{
if(lir_parport_permission()==TRUE)
  {
  allow_parport=TRUE;
  }
else  
  {
  allow_parport=FALSE;
  }
}

void set_hardware_rx_gain(int band_no)
{
int dummy;
dummy=band_no;
}

void set_hardware_rx_frequency(void)
{
}

void clear_hware_data(void)
{
}

void control_hware(void)
{
}
*/
// **********************************************************
//                  WSE drive routines
// **********************************************************
// The radio hardware may be designed in many different ways.
// When the user selects a new frequency there may be several
// different hardware units that have to be changed.
// The routine supplied here is designed to work with the
// RXxxx -> RX70 -> RX10700 -> RX2500 converters.
// If you have some other hardware, you may replace or modify these 
// routines and move them into (w)users_hwaredriver.c   
//
// Associate a number to each frequency band that will
// need a separate case in set_hardware_frequency.
#define AUDIO_MODE 0
#define RX2500_TO_ANT 1
#define RX10700_TO_ANT 2
#define RX70_TO_ANT 3
#define RX144_TO_ANT 4
#define RXHFA_TO_ANT_1800 5
#define RXHFA_TO_ANT_3500 6
#define RXHFA_TO_ANT_7000 7
#define RXHFA_TO_ANT_10000 8
#define RXHFA_TO_ANT_14000 9
#define MAX_BANDS 10

#define TEST_GROUP_Y1 4
#define TEST_GROUP_Y2 TEST_GROUP_Y1 +10
#define TEST_GROUP_Y3 TEST_GROUP_Y2 +8

char *undefmsg={"Undef"};
char *device_name_rx10700={"RX10700"};
char *device_name_rx70={"RX70"};
char *device_name_rx144={"RX144"};
char *device_name_rxhfa={"RXHFA"};
char *device_name_rxhfa_gain={"Gain(RXHFA)"};

char device_bits[256];  
int device_data[256];
unsigned char hware_updates[256];
char *device_name[256];
int hware_retry;
int hware_device_no;
int hware_data;
int hware_bits;
int hware_no_of_updates;
char hware_outdat;


typedef struct {
double increment;
double lowest;
double highest;
int direction;
int label;
} FQ_INFO;

FQ_INFO fqinfo[MAX_BANDS]=
            {0.0  ,   0.0  ,   0.0  , 1, AUDIO_MODE,
             0.025,   1.675,   2.150,-1, RXHFA_TO_ANT_1800,
             0.025,   2.500,   2.500, 1, RX2500_TO_ANT,
             0.025,   3.475,   3.950,-1, RXHFA_TO_ANT_3500,
             0.025,   6.975,   7.450,-1, RXHFA_TO_ANT_7000,
             0.025,   9.975,  10.450,-1, RXHFA_TO_ANT_10000,
             0.025,  10.675,  10.750,-1, RX10700_TO_ANT,
             0.025,  13.975,  14.450,-1, RXHFA_TO_ANT_14000,
             0.025,  69.975,  70.450,-1, RX70_TO_ANT,
             0.025, 143.975, 145.950,-1, RX144_TO_ANT
             }; 
    




void users_close_devices(void)
{
// WSE units do not use the parallel port.
}

void users_open_devices(void)
{
if(ui.parport == 0)
  {
  allow_parport=FALSE;
  }
else
  {
  if(lir_parport_permission()==TRUE)
    {
    allow_parport=TRUE;
    }
  else  
    {
    allow_parport=FALSE;
    }
  }  
}

void hware_interface_test(void)
{
int i, pardat, j, k, xpos, bits[8];
int cntrldat;
char inpins[5]={15,13,12,10,11};
char s[80]; 
clear_screen();
clear_hware_data();
pardat=0;
cntrldat=0;
lir_text(3,TEST_GROUP_Y1,"DATA PORT (address, output only)");
lir_text(2,TEST_GROUP_Y1+2,"Enter 0 to 7 for bit to toggle");
lir_text(5,TEST_GROUP_Y1+4,"Bit:");
lir_text(2,TEST_GROUP_Y1+5,"Status:");
lir_text(5,TEST_GROUP_Y1+6,"Pin:");
lir_text(3,TEST_GROUP_Y3,"CONTROL PORT input pins");
lir_text(5,TEST_GROUP_Y3+3,"Bit:");
lir_text(2,TEST_GROUP_Y3+4,"Status:");
lir_text(5,TEST_GROUP_Y3+5,"Pin:");
lir_text(5,TEST_GROUP_Y2,"Enter D to toggle data (pin 17)");
lir_text(5,TEST_GROUP_Y2+1,"Enter C to toggle clock (pin 1)");
goto show;
loop:;
lir_refresh_screen();
lir_sleep(10000);
test_keyboard();
if(kill_all_flag) return;
if(lir_inkey != 0)
  {
  process_current_lir_inkey();
  if(kill_all_flag) return;
  if(lir_inkey >= '0' && lir_inkey <= '7')
    {
    k=1;
    j=lir_inkey-'0';
    while(j != 0)
      {
      k<<=1;
      j--;
      }
    pardat^=k;
    }
  if(lir_inkey=='D')
    {
    cntrldat ^= HWARE_DATA;
    }
  if(lir_inkey=='C')
    {
    cntrldat ^= HWARE_CLOCK;
    }
show:;  
  j=pardat;
  k=0;
  while(k<8)
    {
    bits[k]=j&1;
    if(bits[k] >0)bits[k]=1;
    j>>=1;
    k++;
    }
  xpos=11;
  for(k=0; k<8; k++)
    {
    sprintf(s,"%1d",k);
    lir_text(xpos,TEST_GROUP_Y1+4,s);
    sprintf(s,"%1d",bits[k]);
    lir_text(xpos,TEST_GROUP_Y1+5,s);
    sprintf(s,"%1d",k+2);
    lir_text(xpos,TEST_GROUP_Y1+6,s);
    xpos+=3;
    }
  for(i=0; i<20;i++)s[i]=' ';
  s[20]=0;  
  lir_text(xpos+5,TEST_GROUP_Y1+5,s);
  lir_text(xpos+5,TEST_GROUP_Y1+5,device_name[pardat]);
  s[1]=0;
  if( (cntrldat&HWARE_DATA) == 0)
    {
    s[0]='H';
    }
  else
    {
    s[0]='L';
    }  
  lir_text(3,TEST_GROUP_Y2+3,"Data=");
  lir_text(8,TEST_GROUP_Y2+3,s);
  if( (cntrldat&HWARE_CLOCK) == 0)
    {
    s[0]='H';
    }
  else
    {
    s[0]='L';
    }  
  lir_text(23,TEST_GROUP_Y2+3,"Clock=");
  lir_text(29,TEST_GROUP_Y2+3,s);
  lir_outb(pardat,ui.parport);
  lir_outb(cntrldat,parport_control);
  }
i=lir_inb(parport_status);
xpos=11;
i >>= 3;
j=8;
for(k=0; k<5; k++)
  {
  settextcolor(7);
  if( (j & parport_ack) !=0)settextcolor(12);
  sprintf(s,"%1d",k+3);
  lir_text(xpos+1,TEST_GROUP_Y3+3,s);
  sprintf(s,"%1d",i&1);
  lir_text(xpos+1,TEST_GROUP_Y3+4,s);
  sprintf(s,"%1d",inpins[k]);
  lir_text(xpos,TEST_GROUP_Y3+5,s);
  xpos+=3;
  i >>= 1;
  j <<= 1;
  }
settextcolor(7);
if(lir_inkey != 'X')goto loop;
}

void set_rxhfa_data(int band)
{
if(device_data[RX144] != band)
  {
  device_data[RXHFA]=band;
  device_bits[RXHFA]=5;
  hware_updates[hware_no_of_updates]=RXHFA;
  hware_no_of_updates++;
  }
}

int set_rx144_data(int stepno)
{
int i;
i=8;
while(stepno >= 20)
  {
  i>>=1;
  stepno-=20;
  }
if(device_data[RX144] != i)
  {
  device_data[RX144]=i;
  device_bits[RX144]=4;
  hware_updates[hware_no_of_updates]=RX144;
  hware_no_of_updates++;
  }
return stepno;  
}

void set_rx10700_data(int stepno)
{
int i;
i=8;
while(stepno != 0)
  {
  i>>=1;
  stepno--;
  }
if(device_data[RX10700] != i)
  {
  device_data[RX10700]=i;
  device_bits[RX10700]=4;
  hware_updates[hware_no_of_updates]=RX10700;
  hware_no_of_updates++;
  }
}

int set_rx70_data(int stepno)
{
int i;
i=16;
while(stepno >= 4)
  {
  i>>=1;
  stepno-=4;
  }
if(device_data[RX70] != i)
  {
  device_data[RX70]=i;
  device_bits[RX70]=5;
  hware_updates[hware_no_of_updates]=RX70;
  hware_no_of_updates++;
  }
return stepno;
}

void set_hardware_rx_gain(int band_no)
{
// bit0 to bit 2 for amplifier/bypass/-10 dB.
// bit0-bit2 = 6 => -10dB
// bit0-bit2 = 3 => 0dB
// bit0-bit2 = 5 => +10dB
//
// bit3 for 5 dB attenuator.
// bit3 = 0 => -5dB
// bit3 = 1 =>  0dB
 
int gaindat[6]={6,14,3,11,5,13};
int i;
if(ui.ad_device_no == SDR14_DEVICE_CODE)
  {
  fg.gain_increment=10;
  fg.gain/=10;
  fg.gain*=10;
  if(fg.gain > 0)fg.gain=0;
  if(fg.gain<-30)fg.gain=-30;
  if(hware_init_flag==0)sdr14_att_counter++;
  return;
  }
if( ui.parport == 0 || !allow_parport)return;
switch (fqinfo[band_no].label)
  {   
  case RXHFA_TO_ANT_1800:
  case RXHFA_TO_ANT_3500:
  case RXHFA_TO_ANT_7000:
  case RXHFA_TO_ANT_10000:
  case RXHFA_TO_ANT_14000:
  fg.gain_increment=5;
  fg.gain/=5;
  fg.gain*=5;
  if(fg.gain < -15)fg.gain=-15;
  if(fg.gain > 10)fg.gain=10;
  i=fg.gain/5+3;
  if(device_data[RXHFA_GAIN] != gaindat[i])
    {
    device_data[RXHFA_GAIN]=gaindat[i];
    device_bits[RXHFA_GAIN]=4;
    if(hware_no_of_updates<0)hware_no_of_updates=0;
    hware_updates[hware_no_of_updates]=RXHFA_GAIN;
    hware_no_of_updates++;
    hware_flag=1;
    hware_time=current_time();
    }
  break;
  
  default:
  break;
  }
}

void set_hardware_rx_frequency(void)
{
float t1;
int band_no, stepno;
if(ui.ad_device_no == SDR14_DEVICE_CODE)
  {
  if(hware_init_flag==0)sdr14_nco_counter++;
  fft1_direction=fg.passband_direction=1;
  fg.passband_increment=.025;
  return;
  }
if( ui.parport == 0 || !allow_parport)return;
// First of all, stop any ongoing activities on the parallel port.
lir_outb(0,ui.parport);
lir_outb(HWARE_CLOCK,parport_control);
// Flag unwritten data as not valid
while(hware_no_of_updates >= 0)
  {
  device_data[hware_updates[hware_no_of_updates]]=-1;
  hware_no_of_updates--;
  }
hware_no_of_updates=0;
hware_flag=0;
band_no=0;
while(band_no < MAX_BANDS-1 && 
     fg.passband_center > fqinfo[band_no].highest)band_no++;
if(fg.passband_center < fqinfo[band_no].lowest && band_no > 0) 
  {
  t1=0.5*(fqinfo[band_no].lowest+fqinfo[band_no-1].highest);
  if(fg.passband_center < t1)band_no--;
  }
if(fg_old_band != band_no)set_hardware_rx_gain(band_no);
fg_old_band=band_no;
t1=(fg.passband_center-fqinfo[band_no].lowest)/fqinfo[band_no].increment;
stepno=t1+0.5;
if(stepno < 0)stepno=0;
set_fq:;
fg.passband_center=fqinfo[band_no].lowest+stepno*fqinfo[band_no].increment;  
if(fg.passband_center > fqinfo[band_no].highest+0.0000001)
  {
  stepno--;
  goto set_fq;
  }  
fg.passband_increment = fqinfo[band_no].increment;
fg.passband_direction = fqinfo[band_no].direction;
fft1_direction=fg.passband_direction;
// Send a message to the screen thread to redraw the frequency scale
// on the main spectrum.
sc[SC_WG_FQ_SCALE]++;
// Set up data to switch relays to connect units properly.
// and to select the correct LO frequency for all units in use.  
switch (fqinfo[band_no].label)
  {   
  case AUDIO_MODE:
// Send nothing to the hardware, just set frequency
// and direction to zero so frequency scale in waterfall and
// other graphs gives audio frequency.
  break;
     
  case RX2500_TO_ANT:
// Route the signal source to the RX2500 input. 
  break;
  
  case RX10700_TO_ANT:
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RX10700 input. 
  goto set_10700;
  
  case RX70_TO_ANT:
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RX10700 input. 
  goto set_70;

  case RXHFA_TO_ANT_1800:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(16);
  goto set_70;

  case RXHFA_TO_ANT_3500:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(8);
  goto set_70;

  case RXHFA_TO_ANT_7000:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(4);
  goto set_70;

  case RXHFA_TO_ANT_10000:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(2);
  goto set_70;

  case RXHFA_TO_ANT_14000:
// Route RXHFA output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RXHFA input. 
  set_rxhfa_data(1);
  goto set_70;

  case RX144_TO_ANT:
// Route RX144 output to the RX70 input. 
// Route RX70 output to the RX10700 input. 
// Route RX10700 output to the RX2500 input. 
// Route the signal source to the RX10700 input. 
  stepno=set_rx144_data(stepno);
set_70:;
  stepno=set_rx70_data(stepno);
set_10700:;
  set_rx10700_data(stepno);
  break;
  }
if(hware_no_of_updates != 0)
  {
  hware_flag=1;
  hware_time=current_time();
  }
else
  {
  hware_init_flag=0;
  }
}

void clear_hware_data(void)
{
int i;
// Set -1 in device_data. 
// A data word is max 31 bit, the sign bit is a flag
// saying the data is not valid.
for(i=0; i<256; i++)
  {
  device_data[i]=-1;
  device_name[i]=undefmsg;
  }
device_name[RX10700]=device_name_rx10700;
device_name[RX70]=device_name_rx70;
device_name[RX144]=device_name_rx144;
device_name[RXHFA]=device_name_rxhfa;
device_name[RXHFA_GAIN]=device_name_rxhfa_gain;
hware_init_flag=1;
hware_no_of_updates=0;
hware_flag=0;
}

void control_hware(void)
{
int err;
unsigned char outdat;
char s[80];
switch(hware_flag)
  {
  case 1:
  hware_no_of_updates--;
  if(hware_no_of_updates <0)
    {
    lirerr(988342);
    return;
    }
  hware_device_no=hware_updates[hware_no_of_updates];
  hware_error_flag=0;
  hware_data=device_data[hware_device_no];
  hware_bits=device_bits[hware_device_no];
// A new unit is selected.
// Write the number of this unit to the data port
  outdat=hware_device_no;
  lir_outb(outdat,ui.parport);
  hware_flag=2;
// Set clock low and make data the first bit of current output.
  if( (hware_data & 1) == 0)
    {
    hware_outdat=HWARE_DATA+HWARE_CLOCK;
    }
  else
    {
    hware_outdat=HWARE_CLOCK;
    }  
  break;

  case 2:
// Set clock high to clock in new data bit.
  hware_outdat^=HWARE_CLOCK;
  hware_flag=3;
  hware_retry=0;
  break;

  case 3:
// Read status to see if we really got the data bit into the unit.
  outdat=lir_inb(parport_status)&parport_ack;
  outdat^=parport_ack_sign;
  err=0;
  if(outdat == 0)
    {
    if( (hware_outdat&HWARE_DATA) != 0)
      {
      err=1;
      }
    }
  else  
    {
    if( (hware_outdat&HWARE_DATA) == 0)
      {
      err=1;
      }
    }
  hware_retry++;
  if(err==1)
    {
    if(hware_retry<4)return;
    hware_error_flag=1;
    }
  hware_bits--;
  if(hware_bits == 0)
    {
// No more. Deselect unit.
    lir_outb(0,ui.parport);
    hware_flag=4;
    return;
    }
  else
    {
// Set clock low and make data the next bit of current output.
    hware_data=hware_data>>1;
    if( (hware_data & 1) == 0)
      {
      hware_outdat=HWARE_DATA+HWARE_CLOCK;
      }
    else
      {
      hware_outdat=HWARE_CLOCK;
      }  
    hware_flag=2;
    }
  break;

  case 4:
// Select unit. This time we actually clock the new data 
// from the shift register into the latch.
  outdat=hware_device_no;
  lir_outb(outdat,ui.parport);
  hware_flag=5;
  return;

  case 5:
  lir_outb(0,ui.parport);
  if(hware_no_of_updates > 0)
    {
    hware_flag=1;
    }
  else
    {  
    hware_flag=0;
    hware_init_flag=0;
    hware_no_of_updates=-1;
    }
  hware_outdat=HWARE_CLOCK;
  if(hware_error_flag != 0)
    {
    sprintf(s,"ERROR: %s",device_name[hware_device_no]);
    wg_error(s,5);
    }

  break;      
  }
lir_outb(hware_outdat,parport_control);
}
// ********************************************************
//               End WSE routines
// ******************************************************** 

