
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "globdef.h"

#if(OSNUM == OS_FLAG_WINDOWS)
void speech_proc_lab(void){}
#endif

#if(OSNUM == OS_FLAG_LINUX)

#include "uidef.h"
#include "sigdef.h"
#include "txdef.h"
#include "screendef.h"
#include "thrdef.h"
#include "keyboard_def.h"
#include "lconf.h"

// allocate 16 megabytes for phonetics (float format).
#define PHONETICS_SIZE (1024*4096)
#define VOICELAB_AMPLITUDE (0.2*0x7fff)
#define LEVEL_METER_X 65

#define MIC_INPUT_LINE 4
#define PRE_GAIN_LINE 5
#define MIC_FILTER_LINE 6
#define MIC_GAIN_LINE 7
#define AF_FILTER_LINE 8
#define AF_GAIN_LINE 9
#define RF_FILTER1_LINE 10
#define RF_GAIN1_LINE 11
#define RF_FILTER2_LINE 12
#define RF_GAIN2_LINE 13
#define RF_FILTER3_LINE 14
#define ALC_LINE 15

#define VLFMT_REAL 1
#define VLFMT_CMPLX_AMPL 2
#define VLT_YSIZ 50

#define MAX_SPEECHPROC_PARMS 15
#define PHONETICS_COLINC 20


void set_vl_defaults(void)
{
vl.noise_level=-10;
vl.wait_time=1.0;
vl.rx_filter_low=250;
vl.rx_filter_high=2450;
vl.rx_filter_slope=0;
vl.pre_gain=0;
vl.pre_agc_time=.05;
vl.mic_filter_low=250;
vl.mic_filter_high=2450;
vl.mic_filter_slope=0;
vl.mic_gain=10;
vl.mic_agc_time=0.02;
vl.af_filter_low=250;
vl.af_filter_high=2450;
vl.af_filter_slope=0;
vl.af_gain=0;
vl.af_agc_time=0.2;
vl.rf_filter1_low=250;
vl.rf_filter1_high=2450;
vl.rf_filter1_slope=0;
vl.rf_gain1=15;
vl.rf_agc_time1=0.01;
vl.rf_filter2_low=250;
vl.rf_filter2_high=2450;
vl.rf_filter2_slope=0;
vl.rf_gain2=0;
vl.rf_agc_time2=0.05;
vl.rf_filter3_low=150;
vl.rf_filter3_high=2550;
vl.rf_filter3_slope=0.;
vl.alc_gain=0;
vl.alc_time=1;

}



void init_vl_trace(int fmt)
{
int i, j, k;
float t1,t2;
// Find the first point of "maximum signal".
t1=0;
i=0;
switch (fmt)
  {
  case VLFMT_REAL:
  for(i=0; i<vl_size; i++)
    {
    if(t1 < fabs(vl_buffer[2*i]))
      {
      t1=fabs(vl_buffer[2*i]);
      vlt_pointer=i;
      }
    }
  vl_begin=0;
  while( fabs(vl_buffer[2*vl_begin]) <0.25*t1)vl_begin++;
  vl_end=vl_size-1;
  while( fabs(vl_buffer[2*vl_end]) <0.25*t1)vl_end--;
  break;
  
  case VLFMT_CMPLX_AMPL:
  for(i=0; i<vl_size; i++)
    {
    t2=pow(vl_buffer[2*i],2.0)+pow(vl_buffer[2*i+1],2.0);
    if(t1 < t2)
      {
      t1=t2;
      vlt_pointer=i;
      }
    }
  vl_begin=0;
  while( pow(vl_buffer[2*vl_begin],2.0)+
                     pow(vl_buffer[2*vl_begin+1],2.0) < 0.0625*t1)vl_begin++;
  vl_end=vl_size-1;
  while( pow(vl_buffer[2*vl_end],2.0)+
                     pow(vl_buffer[2*vl_end+1],2.0)<0.0625*t1)vl_end--;
  t1=sqrt(t1);
  break;
  }
if(screen_width*vlt_step > vl_last_pos-vl_first_pos)
  {
  vlt_step=(vl_last_pos-vl_first_pos)/screen_width; 
  }
vlt_pointer-=(0.25*screen_width+vlt_shift)*vlt_step;  
if(vlt_pointer<vl_first_pos)vlt_pointer=vl_first_pos;
if(vlt_pointer+screen_width*vlt_step>vl_last_pos)
  {
  vlt_pointer=vl_last_pos-screen_width*vlt_step;
  }
vlt_ampfac=VLT_YSIZ/t1;
for(i=0; i<MAX_VL_TRACES; i++)
  {
  for(j=1; j<screen_width; j++)
    {
    lir_line(j-1,vl_trace[i*screen_width+j-1],j,vl_trace[i*screen_width+j],0);
    if(kill_all_flag) return;
    }
  }    
j=(screen_height-(ALC_LINE+2.5)*text_height)/MAX_VL_TRACES;
k=screen_height-VLT_YSIZ-2*text_height;
for(i=0; i<MAX_VL_TRACES; i++)
  {
  vlt_y0[i]=k;
  lir_line(0,vlt_y0[i]-VLT_YSIZ,screen_width-1,vlt_y0[i]-VLT_YSIZ,0);  
  if(kill_all_flag) return;
  lir_line(0,vlt_y0[i]+VLT_YSIZ,screen_width-1,vlt_y0[i]+VLT_YSIZ,0);  
  if(kill_all_flag) return;
  k-=j;
  }
}  


void draw_vl_trace(int fmt, int traceno)
{
int i,j;
short int *z;
float *x;
float t1,t2,t3;
if(fixed_phonetics_flag == 0 && vl_signal_type == 0)return;
if(traceno >=MAX_VL_TRACES)
  {
  lirerr(187295);
  return;
  }
z=&vl_trace[traceno*screen_width];
x=&vl_buffer[2*vlt_pointer];
switch (fmt)
  {
  case VLFMT_REAL:
  for(i=0; i<screen_width; i+=2*text_width)
    {
    lir_line(i,vlt_y0[traceno]-VLT_YSIZ,i+text_width/2,
                                      vlt_y0[traceno]-VLT_YSIZ,8);
    if(kill_all_flag) return;
    }
  for(i=text_width; i<screen_width; i+=2*text_width)
    {
    lir_line(i,vlt_y0[traceno]+VLT_YSIZ,i+text_width/2,
                                      vlt_y0[traceno]+VLT_YSIZ,8);
    if(kill_all_flag) return;
    }
  t3=0;
  for(i=0; i<screen_width; i++)
    {
    t1=-BIG;
    t2=BIG;
    for(j=0;j<vlt_step;j++)
      {
      if(t1 < x[2*(i*vlt_step+j)])
        {
        t1=x[2*(i*vlt_step+j)];
        }
      if(t2 > x[2*(i*vlt_step+j)])
        {
        t2=x[2*(i*vlt_step+j)];
        }
      }
    if(i==0)
      {
      if(fabs(t2) > fabs(t1))t1=t2;
      }
    else
      {
      if(t1*t2 < 0)
        {
        if(fabs(t2-t3) > fabs(t1-t3))t1=t2;
        }
      else
        {
        if(fabs(t2) > fabs(t1))t1=t2;
        }
      }    
    t3=t1;
    z[i]=vlt_y0[traceno]-t1*vlt_ampfac;
    }
  for(i=1; i<screen_width; i++)
    {
    lir_line(i-1,z[i-1],i,z[i],15);
    if(kill_all_flag) return;
    }
  break;

  case VLFMT_CMPLX_AMPL:
  lir_line(0,vlt_y0[traceno],screen_width-1,vlt_y0[traceno],8);  
  if(kill_all_flag) return;
  for(i=0; i<screen_width; i+=2*text_width)
    {
    lir_line(i,vlt_y0[traceno]-VLT_YSIZ,i+text_width/2,
                                                vlt_y0[traceno]-VLT_YSIZ,8);
    if(kill_all_flag) return;
    }
  for(i=0; i<screen_width; i++)
    {
    t1=0;
    for(j=0;j<vlt_step;j++)
      {
      t2=pow(x[2*(i*vlt_step+j)],2.0)+pow(x[2*(i*vlt_step+j)+1],2.0);
      if(t1 < t2)t1=t2;
      }
    z[i]=vlt_y0[traceno]-sqrt(t1)*vlt_ampfac;
    }
  for(i=1; i<screen_width; i++)
    {
    lir_line(i-1,z[i-1],i,z[i],15);
    if(kill_all_flag) return;
    }
  break;
  }
}


float noisegen(void) 
{
// Return a number distributed following a gaussian
// Mean value is 0 and sigma is vl.noise_level with respect to
// 10% of the clipping level of the D/A converter 
float x, y, z; 
y = lir_random(); 
z = lir_random(); 
x = z * 2*PI_L; 
return VOICELAB_AMPLITUDE*sin(x)*sqrt(-2*log(y))*pow(10.,0.05*vl.noise_level);
}


void show_speechproc_parm(int parno)
{
char s[256];
switch (parno)
  {
  case 1:
  sprintf(s,"01: Noise level:  %5.2f dB ",vl.noise_level);
  lir_text(0,2,s);
  break;
  
  case 2:
  sprintf(s,"02: Wait time:    %5.2f s ",vl.wait_time);
  lir_text(40,2,s);
  break;

  case 3:
  sprintf(s,"03: Rx filter:    %4.0f to %4.0f Hz. Slope %5.2f dB/kHz. ",
                       vl.rx_filter_low,vl.rx_filter_high,vl.rx_filter_slope);
  lir_text(0,3,s);
  break;

  case 4:
  sprintf(s,"04: Pre gain:     %5.2f dB. T=%.3f sec. ",
                                                vl.pre_gain,vl.pre_agc_time);
  lir_text(0,PRE_GAIN_LINE,s);
  break;


  case 5:
  sprintf(s,"05: Mic filter:   %4.0f to %4.0f Hz. Slope %5.2f dB/kHz. ",
                   vl.mic_filter_low,vl.mic_filter_high,vl.mic_filter_slope);
  lir_text(0,MIC_FILTER_LINE,s);
  break;

  case 6:
  sprintf(s,"06: Mic gain:     %5.2f dB. T=%.3f sec. ",
                                               vl.mic_gain,vl.mic_agc_time);
  lir_text(0,MIC_GAIN_LINE,s);
  break;

  case 7:
  sprintf(s,"07: AF filter:    %4.0f to %4.0f Hz. Slope %5.2f dB/kHz. ",
                   vl.af_filter_low,vl.af_filter_high,vl.af_filter_slope);
  lir_text(0,AF_FILTER_LINE,s);
  break;

  case 8:
  sprintf(s,"08: AF gain:      %5.2f dB. T=%.3f sec. ",
                                               vl.af_gain,vl.af_agc_time);
  lir_text(0,AF_GAIN_LINE,s);
  break;

  case 9:
  sprintf(s,"09: RF filter 1:  %4.0f to %4.0f Hz. Slope %5.2f dB/kHz. ",
                   vl.rf_filter1_low,vl.rf_filter1_high,vl.rf_filter1_slope);
  lir_text(0,RF_FILTER1_LINE,s);
  break;

  case 10:
  sprintf(s,"10: RF gain 1:    %5.2f dB. T=%.3f sec. ",
                                                vl.rf_gain1,vl.rf_agc_time1);
  lir_text(0,RF_GAIN1_LINE,s);
  break;

  case 11:
  sprintf(s,"11: RF filter 2:  %4.0f to %4.0f Hz. Slope %5.2f dB/kHz. ",
                   vl.rf_filter2_low,vl.rf_filter2_high,vl.rf_filter2_slope);
  lir_text(0,RF_FILTER2_LINE,s);
  break;

  case 12:
  sprintf(s,"12: RF gain 2:    %5.2f dB. T=%.3f sec. ",
                                               vl.rf_gain2,vl.rf_agc_time2);
  lir_text(0,RF_GAIN2_LINE,s);
  break;

  case 13:
  sprintf(s,"13: RF filter 3:  %4.0f to %4.0f Hz. Slope %5.2f dB/kHz. ",
                   vl.rf_filter3_low,vl.rf_filter3_high,vl.rf_filter3_slope);
  lir_text(0,RF_FILTER3_LINE,s);
  break;

  case 14:
  sprintf(s,"14: ALC gain:     %5.2f dB. T=%.3f sec. ",
                                                 vl.alc_gain,vl.alc_time);
  lir_text(0,ALC_LINE,s);
  break;

  }

}  

float stdparm(float min, float max)
{
clear_lines(screen_last_line-1,screen_last_line);
lir_text(0,screen_last_line,"New value: >");
return lir_get_float(13, screen_last_line, 15, min, max);
}

void change_speechproc_parm(int parno)
{
settextcolor(12);
show_speechproc_parm(parno);
settextcolor(15);
lir_text(0,screen_last_line,"New value: >");
switch (parno)
  {
  case 1:
  vl.noise_level=stdparm(-99.,99.);
  break;
  
  case 2:
  vl.wait_time=stdparm(0.,10.);
  break;

  case 3:
  lir_text(0,screen_last_line,"Low:");
  vl.rx_filter_low=lir_get_float(4, screen_last_line, 9, 20., 1500);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"High:");
  vl.rx_filter_high=lir_get_float(5, screen_last_line, 9, 
                             vl.rx_filter_low+20, tx.input_speed/2);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"Slope:");
  vl.rx_filter_slope=lir_get_float(6,screen_last_line,9, -30. ,30.);
  break;

  case 4:
  lir_text(0,screen_last_line,"Gain:");
  vl.pre_gain=lir_get_float(5, screen_last_line, 9, -20., 40);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"AGC time:");
  vl.pre_agc_time=lir_get_float(9, screen_last_line, 9, -10., 5.);
  break;

  case 5:
  lir_text(0,screen_last_line,"Low:");
  vl.mic_filter_low=lir_get_float(4, screen_last_line, 9, 20., 1500);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"High:");
  vl.mic_filter_high=lir_get_float(5, screen_last_line, 9, 
                             vl.rx_filter_low+20, tx.input_speed/2);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"Slope:");
  vl.mic_filter_slope=lir_get_float(6,screen_last_line,9, -30. ,30.);
  break;

  case 6:
  lir_text(0,screen_last_line,"Gain:");
  vl.mic_gain=lir_get_float(5, screen_last_line, 9, -20., 40);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"AGC time:");
  vl.mic_agc_time=lir_get_float(9, screen_last_line, 9, -10., 5.);
  break;

  case 7:
  lir_text(0,screen_last_line,"Low:");
  vl.af_filter_low=lir_get_float(4, screen_last_line, 9, 20., 1500);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"High:");
  vl.af_filter_high=lir_get_float(5, screen_last_line, 9, 
                             vl.af_filter_low+20, tx.input_speed/2);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"Slope:");
  vl.af_filter_slope=lir_get_float(6,screen_last_line,9, -30. ,30.);
  break;

  case 8:
  lir_text(0,screen_last_line,"Gain:");
  vl.af_gain=lir_get_float(5, screen_last_line, 9, -20., 40);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"AGC time:");
  vl.af_agc_time=lir_get_float(9, screen_last_line, 9, -10., 5.);
  break;

  case 9:
  lir_text(0,screen_last_line,"Low:");
  vl.rf_filter1_low=lir_get_float(4, screen_last_line, 9, 20., 1500);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"High:");
  vl.rf_filter1_high=lir_get_float(5, screen_last_line, 9, 
                             vl.rf_filter1_low+20, tx.input_speed/2);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"Slope:");
  vl.rf_filter1_slope=lir_get_float(6,screen_last_line,9, -30. ,30.);
  break;

  case 10:
  lir_text(0,screen_last_line,"Gain:");
  vl.rf_gain1=lir_get_float(5, screen_last_line, 9, -20., 60);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"AGC time:");
  vl.rf_agc_time1=lir_get_float(9, screen_last_line, 9, -10., 5.);
  break;

  case 11:
  lir_text(0,screen_last_line,"Low:");
  vl.rf_filter2_low=lir_get_float(4, screen_last_line, 9, 20., 1500);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"High:");
  vl.rf_filter2_high=lir_get_float(5, screen_last_line, 9, 
                             vl.rf_filter2_low+20, tx.input_speed/2);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"Slope:");
  vl.rf_filter2_slope=lir_get_float(6,screen_last_line,9, -30. ,30.);
  break;

  case 12:
  lir_text(0,screen_last_line,"Gain:");
  vl.rf_gain2=lir_get_float(5, screen_last_line, 9, -20., 60);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"AGC time:");
  vl.rf_agc_time2=lir_get_float(9, screen_last_line, 9, -10., 5.);
  break;

  case 13:
  lir_text(0,screen_last_line,"Low:");
  vl.rf_filter3_low=lir_get_float(4, screen_last_line, 9, 20., 1500);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"High:");
  vl.rf_filter3_high=lir_get_float(5, screen_last_line, 9, 
                             vl.rf_filter3_low+20, tx.input_speed/2);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"Slope:");
  vl.rf_filter3_slope=lir_get_float(6,screen_last_line,9, -30. ,30.);
  break;

  case 14:
  lir_text(0,screen_last_line,"Gain:");
  vl.alc_gain=lir_get_float(5, screen_last_line, 9, -20., 40);
  if(kill_all_flag) return;
  lir_text(0,screen_last_line,"ALC time:");
  vl.alc_time=lir_get_float(9, screen_last_line, 9, -10., 5.);
  break;
  }
if(kill_all_flag) return;
settextcolor(7);
show_speechproc_parm(parno);
clear_lines(screen_last_line,screen_last_line);
}


void vl_level_meter(int line, int type)
{
char s[80];
int i;
float t1,t2,t3;
if(type == 1)
  {
  t1=0;
  for(i=vl_begin; i<=vl_end; i++)
    {
    t1+=pow(vl_buffer[2*i],2.);
    }
  t1/=vl_end-vl_begin+1;
  t2=0;
  for(i=0; i<vl_size; i++)
    {
    if( t2 < fabs(vl_buffer[2*i]))
      {
      t2=fabs(vl_buffer[2*i]);
      }
    }
  t2=t2*t2;
  sprintf(s," Peak %6.2f dB  Avg %6.2f dB (Real)    ",10*log10(t2),10*log10(t1));
  }
else
  {
  t1=0;
  for(i=vl_begin; i<=vl_end; i++)
    {
    t3=vl_buffer[2*i]*vl_buffer[2*i]+vl_buffer[2*i+1]*vl_buffer[2*i+1];
    t1+=t3;
    }
  t1/=vl_end-vl_begin+1;
  t2=0;
  for(i=0; i<vl_size; i++)
    {
    t3=vl_buffer[2*i]*vl_buffer[2*i]+vl_buffer[2*i+1]*vl_buffer[2*i+1];
    if( t2 < t3)
      {
      t2=t3;
      }
    }
  sprintf(s," Peak %6.2f dB  Avg %6.2f dB (Complex) ",10*log10(t2),10*log10(t1));
  }
settextcolor(15);
switch (vl_traces_mode)
  {
  case 0:
  switch (line)
    {
    case MIC_INPUT_LINE:
    strcat(s,"trk[0]");
    break;
    
    case MIC_GAIN_LINE:
    strcat(s,"trk[1]");
    break;
    
    case AF_GAIN_LINE:
    strcat(s,"trk[2]");
    break;
    
    case RF_GAIN1_LINE:
    strcat(s,"trk[3]");
    break;
    
    case ALC_LINE:
    strcat(s,"trk[4]");
    break;
    
    default:
    strcat(s,"      ");
    break;
    }
  break;

  case 1:
  switch (line)
    {
    case MIC_GAIN_LINE:
    strcat(s,"trk[0]");
    break;
    
    case AF_FILTER_LINE:
    strcat(s,"trk[1]");
    break;
    
    case RF_GAIN1_LINE:
    strcat(s,"trk[2]");
    break;

    case ALC_LINE:
    strcat(s,"trk[3]");
    break;
    
    default:
    strcat(s,"      ");
    break;
    }
  break;

  case 2:
  switch (line)
    {
    case AF_GAIN_LINE:
    strcat(s,"trk[0]");
    break;
  
    case RF_GAIN1_LINE:
    strcat(s,"trk[1]");
    break;
  
    case RF_GAIN2_LINE:
    strcat(s,"trk[2]");
    break;
  
    case ALC_LINE:
    strcat(s,"trk[3]");
    break;
    
    default:
    strcat(s,"      ");
    break;
    }
  break;
  
  default:
  strcat(s,"      ");
  break;
  }

lir_text(LEVEL_METER_X,line,s);
}

void vl_gain_and_agc(float gain, float agc_time, int fmtflag)
{
float t1, t2, t3, r1;
int i;
t1=pow(10.,0.05*gain);
// If agc_time is zero or negative (as seen on the screen),
// we use a clipper, otherwise we use an AGC with immediate
// attack and a fall time given by agc time.
// If agc_time is zero, just limit the amplitude to 1.000
// For negative agc_time, use a soft clipper function
// that goes towards 1 as the signal goes to infinity.
// Negative values close to zero give a hard limiter,
// more negative gives soft clipping. 
if(agc_time < 0.000500001)
  {
  if(agc_time < 0)
    {
// Use soft clipping.    
    t2=agc_time;
    if(t2>-.04)t2=-.04;
    t2=1-1/t2;
    if(fmtflag==1)
      {
      for(i=0; i<vl_size; i++)
        {
        vl_buffer[2*i  ]*=t1;
        vl_buffer[2*i+1]=0;
        t3=pow(1/(pow(fabs(vl_buffer[2*i  ]),t2)+1),1/t2);
        vl_buffer[2*i  ]*=t3;
        }
      }
    else
      {    
      for(i=0; i<vl_size; i++)
        {
        vl_buffer[2*i  ]*=t1;
        vl_buffer[2*i+1]*=t1;
        t3=sqrt(vl_buffer[2*i]*vl_buffer[2*i]+vl_buffer[2*i+1]*vl_buffer[2*i+1]);
        t3=pow(1/(pow(t3,t2)+1),1/t2);
        vl_buffer[2*i  ]*=t3;
        vl_buffer[2*i+1]*=t3;
        }
      }
    }
  else
    {  
// agc_time is below 0.001 seconds (as seen on screen). Use a simple clipper.
    if(fmtflag==1)
      {
      for(i=0; i<vl_size; i++)
        {
        vl_buffer[2*i  ]*=t1;
        vl_buffer[2*i+1]=0;
        t3=fabs(vl_buffer[2*i  ]);
        if(t3 > 1)vl_buffer[2*i  ]/=fabs(vl_buffer[2*i  ]);
        }
      }
    else
      {    
      for(i=0; i<vl_size; i++)
        {
        vl_buffer[2*i  ]*=t1;
        vl_buffer[2*i+1]*=t1;
        t3=sqrt(vl_buffer[2*i]*vl_buffer[2*i]+vl_buffer[2*i+1]*vl_buffer[2*i+1]);
        if(t3 > 1)
          {
          vl_buffer[2*i  ]/=t3;
          vl_buffer[2*i+1]/=t3;
          }
        }
      }
    }
  }
else  
  {
// We want the agc to decay towards 1.0 by 1/e (=0.368) in r1 samples
// in the forward time direction
  r1=agc_time*tx.input_speed;
  r1=pow(0.368,1/r1); 
  t2=1.0;
  if(fmtflag==1)
    {
    for(i=0; i<vl_size; i++)
      {
      vl_buffer[2*i  ]*=t1;
      vl_buffer[2*i+1]=0;
      t3=fabs(vl_buffer[2*i  ]);
      if(t3 > t2)
        {
        t2=t3;
        }
      vl_agc[i]=t2;
      t2=t2-1;
      t2*=r1;
      t2=t2+1;
      }
// We want the agc to decay towards 1.0 by 1/e (=0.368) in r1 samples
// in the backward time direction. 
// The rise time of the agc is maximised to 10 milliseconds (100Hz) for
// the modulation introduced to be essentially within the passband.
    if(agc_time > 0.01)
      {
      r1=0.01*tx.input_speed;
      r1=pow(0.368,1/r1); 
      }
    t2=1.0;
    for(i=vl_size-1; i>=0; i--)
      {
      t3=fabs(vl_buffer[2*i  ]);
      if(t3 > t2)
        {
        t2=t3;
        }
      if(t2 > vl_agc[i])
        {
        vl_buffer[2*i  ]/=t2;
        }
      else
        {
        vl_buffer[2*i  ]/=vl_agc[i];
        }                 
      t2=t2-1;
      t2*=r1;
      t2=t2+1;
      }
    }
  else
    {    
    for(i=0; i<vl_size; i++)
      {
      vl_buffer[2*i  ]*=t1;
      vl_buffer[2*i+1]*=t1;
      t3=sqrt(vl_buffer[2*i]*vl_buffer[2*i]+vl_buffer[2*i+1]*vl_buffer[2*i+1]);
      if(t3 > t2)
        {
        t2=t3;
        }
      vl_agc[i]=t2;
      t2=t2-1;
      t2*=r1;
      t2=t2+1;
      }
    if(agc_time > 0.01)
      {
      r1=0.01*tx.input_speed;
      r1=pow(0.368,1/r1); 
      }
    t2=1.0;
    for(i=vl_size-1; i>=0; i--)
      {
      t3=sqrt(vl_buffer[2*i]*vl_buffer[2*i]+vl_buffer[2*i+1]*vl_buffer[2*i+1]);
      if(t3 > t2)
        {
        t2=t3;
        }
      if(t2 > vl_agc[i])
        {
        vl_buffer[2*i  ]/=t2;
        vl_buffer[2*i+1]/=t2;
        }
      else
        {
        vl_buffer[2*i  ]/=vl_agc[i];
        vl_buffer[2*i+1]/=vl_agc[i];
        }                 
      t2=t2-1;
      t2*=r1;
      t2=t2+1;
      }
    }
  }
}

void vl_filter(float low,float high,float slope, int fmtflag, int gain)
{
int i, j, k;
int ia, ib;
float t1,t2,t3,r1;
big_fftforward(vl_size, vl_fftn, vl_buffer, vl_fft_table, vl_fft_permute);
// Clear points outside the selected passband
// The spectrum has positive frequencies in points 0 to vl_size/2-1
// and negative frequencies from vl_size/2 to vl_size-1  
// With complex data the negative frequencies are zero, if
// we have real data they exist.
// Clear negative frequencies (we loose 6 dB which is compensated for
// by the calling routine if needed)
j=vl_size*low/tx.input_speed;
k=vl_size*high/tx.input_speed;
// Find out how many bins there are for 1 kHz.
// set a factor that will give the desired slope.
r1=(float)(gain)/vl_size;
t1=(float)(vl_size*2000)/tx.input_speed;
t2=pow(10.,-slope*0.05);
t1=pow(t2,1/t1);
t2=r1*t1;
t3=r1/t1;
ia=(j+k)/2;
vl_buffer[2*ia  ]*=r1;
vl_buffer[2*ia+1]*=r1;
ia--;
ib=ia+2;
while(ib<k)
  {
  vl_buffer[2*ib]*=t3;
  vl_buffer[2*ib+1]*=t3;
  t3/=t1;
  ib++;
  }
while(ib<vl_size)
  {
  vl_buffer[2*ib]=0;
  vl_buffer[2*ib+1]=0;
  ib++;
  }
while(ia>=j)
  {
  vl_buffer[2*ia]*=t2;
  vl_buffer[2*ia+1]*=t2;
  t2*=t1;
  ia--;
  }
while(ia >= 0)
  {
  vl_buffer[2*ia]=0;
  vl_buffer[2*ia+1]=0;
  ia--;
  } 
big_fftback(vl_size, vl_fftn, vl_buffer, vl_fft_table, vl_fft_permute);
if(fmtflag==1)for(i=0;i<vl_size;i++)vl_buffer[2*i+1]=0;
}

void show_guess_statistics(void)
{
char s[80];
float t1;
t1=vl_true_guesses+vl_false_guesses+vl_missed_guesses;
if(t1 != 0)
  {
  t1=100*vl_true_guesses/t1;
  }
sprintf(s,"True %d   False %d   Missed %d   Total %d       Copy %.1f%% ",
                     vl_true_guesses,vl_false_guesses,vl_missed_guesses,
                     vl_true_guesses+vl_false_guesses+vl_missed_guesses,t1);   
lir_text(0,screen_last_line,s);
}


void run_speech_processing(void)
{
char s[512];
audio_buf_info ad_info;
short int *wrbuf;
float *filtered_output, *raw_output, *rxbuf;
float *rx_fft_win;
float t1,t2,t3,t4;
int i, j, k, ia, ib, wrsamps, waitsamps;
int mask, raw_output_pa, filtered_output_pa;
int raw_output_px,filtered_output_px;
int *voicelab_handle;
int vlstat;
int rx_fftsize, rx_fftn;
COSIN_TABLE *rx_fft_table;
unsigned short int *rx_fft_permute;
FILE* file;
float max_output;
unsigned char current_phonetics, guess_flag;
init_txmem();
lir_srandom();
set_vl_defaults();
vl_minval=0.005;
vlt_step=1;
vlt_shift=0;
vl_traces_mode=0;
vl_signal_type=0;
rx_daout_bytes=ui.max_output_bytes;
if(rx_daout_bytes!=2)lirerr(881004);
open_rx_sndout();
if(kill_all_flag) return;
rx_fftsize=rx_daout_block;
wrsamps=rx_daout_block/2;
vl_size=tx.input_speed*2;
make_power_of_two(&vl_size);
mask=vl_size-1;
init_memalloc(voicelabmem, MAX_VOICELAB_ARRAYS);
mem(1,&wrbuf,rx_daout_channels*rx_daout_block*sizeof(float),0);
mem(2,&filtered_output,vl_size*sizeof(float),0);
mem(3,&raw_output,vl_size*sizeof(float),0);
mem(4,&vl_buffer,2*vl_size*sizeof(float),0);
mem(5,&rxbuf,2*rx_fftsize*sizeof(float),0);
mem(6,&rx_fft_table,rx_fftsize*sizeof(COSIN_TABLE)/2,0);
mem(7,&rx_fft_permute,rx_fftsize*sizeof(short int),0);
mem(8,&rx_fft_win,(1+rx_fftsize/2)*sizeof(float),0);
mem(9,&vl_fft_table,vl_size*sizeof(COSIN_TABLE)/2,0);
mem(10,&vl_fft_permute,vl_size*sizeof(int),0);
mem(11,&vl_trace,MAX_VL_TRACES*screen_width*sizeof(short int),0);
mem(12,&vl_agc,vl_size*sizeof(float),0);
memalloc((int*)(&voicelab_handle),"voicelab");
rx_fftn=0;
i=rx_fftsize;
while(i>1)
  {
  i>>=1;
  rx_fftn++;
  }
init_fft(0, rx_fftn, rx_fftsize, rx_fft_table, rx_fft_permute);
make_window(2,rx_fftsize/2, 2, rx_fft_win);
vl_fftn=0;
i=vl_size;
while(i>1)
  {
  i>>=1;
  vl_fftn++;
  }
init_big_fft(0, vl_fftn, vl_size, vl_fft_table, vl_fft_permute);
clear_screen();
// Simulate transmission path from voice at tx location to
// loudspeaker at the rx location.
// Allow the user to set filters, clippers and other things that
// may affect intelligibility.
settextcolor(14);
lir_text(0,0,"Press correct key if you can hear Phonetics.");
settextcolor(11);
lir_text(47,0,"[+-]=X scale.  [<>]=X pos.  [/]=.GIF file.");
sprintf(s,"[^]=Display mode.  [%c]=Wave mode.",'"');
lir_text(91,0,s);
lir_text(0,1,"[?]=Change parm.  [!]=Clear score. [*]=Exit.");
lir_text(45,1,"[:]=Save parms.  [&]=Recall parms.  [#]=Set Phonetics.");
lir_text(100,1,"[,]=Pause.");
fixed_phonetics_flag=0;
sp_loop_init_a:;
settextcolor(7);
for(i=1; i<MAX_SPEECHPROC_PARMS; i++)show_speechproc_parm(i);
sp_loop_init_b:;
for(i=0; i<=screen_last_col; i++)
  {
  s[i]='-';
  }
s[screen_last_col+1]=0;
settextcolor(14);  
lir_text(0,4,s);
settextcolor(7);
current_phonetics=0;
waitsamps=0;
raw_output_px=0;
filtered_output_px=wrsamps;
raw_output_pa=1;
filtered_output_pa=0;
vlstat=0;
guess_flag=0;
vl_true_guesses=0;
vl_false_guesses=0;
vl_missed_guesses=0;
max_output=0;
for(i=0; i<vl_size; i++)
  {
  raw_output[i]=0;
  filtered_output[i]=0;
  }
sp_loop:;
// Use vlstat to split processing in several steps.
// We may use slow FFTs to do filtering so the complete
// processing of a character may cause underrun errors if
// it is not split in several processing tasks.
switch (vlstat)
  {
// Start the cycle by getting a random Phonetics.  
// Move it into vl_buffer
// NOTE. We use complex FFTs because they are standard within
// Linrad. For real processing we store the data in the real part
// only and make the imaginary part zero. 
  case 0:
  switch (vl_signal_type)
    {
    case 0:
    if(fixed_phonetics_flag != 0)
      {
      current_phonetics=fixed_phonetics_flag;
      sprintf(s,"   %c   ",fixed_phonetics_flag);
      }
    else
      {  
      show_guess_statistics();
      clear_lines(screen_last_line-1,screen_last_line-1);
      sprintf(s,"       ");
      current_phonetics=0;
      while( (phonetics_flag[(int)(current_phonetics)]&2) == 0)
        {
        current_phonetics=(int)(400*lir_random())%256; 
        }
      }
    lir_text(LEVEL_METER_X-7,4,s);
    ia=phonetics_pointer[(int)(current_phonetics)];
    ib=phonetics_pointer[current_phonetics+1];
    phonetics_length=ib-ia;
    j=(vl_size-ib+ia)/2;
    for(i=0; i<j; i++)
      {
      vl_buffer[2*i  ]=0;
      vl_buffer[2*i+1]=0;
      }
    vl_first_pos=j;  
    while(ia!=ib)
      {
      vl_buffer[2*j  ]=phonetics[ia];
      vl_buffer[2*j+1]=0;
      j++;
      ia++;
      }
    vl_last_pos=j-1;  
    while(j<vl_size)
      {
      vl_buffer[2*j  ]=0;  
      vl_buffer[2*j+1]=0;  
      j++;
      }
    break;
    
    case 1:
    sprintf(s," sine  ");
    goto sinsqr;

    case 3:
    sprintf(s,"square ");
    t2=2*PI_L*100/tx.input_speed;
    goto sinsqr;

    case 4:
    sprintf(s,"2-tone ");
 sinsqr:;
    lir_text(LEVEL_METER_X-7,4,s);
    ia=vl_size*.45;
    ib=vl_size*.55;
    phonetics_length=ib-ia;
    j=(vl_size-ib+ia)/2;
    for(i=0; i<j; i++)
      {
      vl_buffer[2*i  ]=0;
      vl_buffer[2*i+1]=0;
      }
    vl_first_pos=j;  
    t1=0;
    switch (vl_signal_type)
      {
      case 1:       
      t2=2*PI_L*600/tx.input_speed;
      while(ia!=ib)
        {
        vl_buffer[2*j  ]=sin(t1);
        t1+=t2;
        vl_buffer[2*j+1]=0;
        j++;
        ia++;
        }
      break;

      case 3:       
      t2=2*PI_L*100/tx.input_speed;
      while(ia!=ib)
        {
        if(sin(t1) >= 0)
          {
          vl_buffer[2*j  ]=1;
          }
        else
          {
          vl_buffer[2*j  ]=-1;
          }
        t1+=t2;
        vl_buffer[2*j+1]=0;
        j++;
        ia++;
        }
      break;

      case 4:
      t3=0;       
      t2=2*PI_L*1000/tx.input_speed;
      t4=2*PI_L*1225/tx.input_speed;
      while(ia!=ib)
        {
        vl_buffer[2*j  ]=0.5*(sin(t1)+sin(t3));
        t1+=t2;
        t3+=t4;
        vl_buffer[2*j+1]=0;
        j++;
        ia++;
        }
      break;
      }
    vl_last_pos=j;  
    while(j<vl_size)
      {
      vl_buffer[2*j  ]=0;  
      vl_buffer[2*j+1]=0;  
      j++;
      }
    k=vl_first_pos;
    vl_first_pos*=0.95;
    j=vl_last_pos;  
    vl_last_pos*=1.05;  
    t2=2*PI_L*10/tx.input_speed;
    t1=0;
    while(t1<0.5*PI_L)
      {
      j--;
      vl_buffer[2*k]*=pow(sin(t1),2.0);
      vl_buffer[2*j]*=pow(sin(t1),2.0);
      t1+=t2;
      k++;
      }
    break;
    
    case 2:
    sprintf(s," pulse ");
    lir_text(LEVEL_METER_X-7,4,s);
    ia=vl_size*.45;
    ib=vl_size*.55;
    phonetics_length=ib-ia;
    j=(vl_size-ib+ia)/2;
    for(i=0; i<j; i++)
      {
      vl_buffer[2*i  ]=0;
      vl_buffer[2*i+1]=0;
      }
    vl_first_pos=j*0.97;  
    t1=0;
    t2=2*PI_L*300/tx.input_speed;
    i=0;
    while(ia!=ib)
      {
      if(i%6==0 && t1>=0)
        {
        vl_buffer[2*j  ]=exp(-pow(4*(t1-PI_L/2),2.0));
        }
      else
        {
        vl_buffer[2*j  ]=0;
        }
      t1+=t2;
      if(t1 > PI_L)
        {
        i++;
        t1-=2*PI_L;
        }
      vl_buffer[2*j+1]=0;
      j++;
      ia++;
      }
    vl_last_pos=j*1.03;  
    while(j<vl_size)
      {
      vl_buffer[2*j  ]=0;  
      vl_buffer[2*j+1]=0;  
      j++;
      }
    break;
    } 
  init_vl_trace(VLFMT_REAL);
  if(kill_all_flag) goto run_speech_x;
  vl_level_meter(MIC_INPUT_LINE,1);
  if(vl_signal_type == 0)
    {
    t1=vl_minval*vlt_ampfac/VLT_YSIZ;
    for(j=vl_first_pos; j<vl_last_pos; j++)
      {
      if( fabs(vl_buffer[2*j-2]) < t1 &&
          fabs(vl_buffer[2*j  ]) < t1 &&
          fabs(vl_buffer[2*j+2]) < t1)
        {
        vl_buffer[2*j  ]=0;
        }
      }    
    }
  if(vl_traces_mode==0)
    {
    draw_vl_trace(VLFMT_REAL,0);
    if(kill_all_flag) goto run_speech_x;
    }
  vlstat=91;
  break;

  case 91:
  vl_gain_and_agc(vl.pre_gain,vl.pre_agc_time,1);
  vlstat=1;
  vlt_ampfac=VLT_YSIZ;
  vl_level_meter(PRE_GAIN_LINE,1);
  break;

  case 1:
// First processing step.
// Filter the signal through the microphone filter
  vl_filter(vl.mic_filter_low,vl.mic_filter_high,vl.mic_filter_slope,1,2);
  vl_level_meter(MIC_FILTER_LINE,1);
  vlstat=2;
  break;

  case 2:
// Second processing step.
// Apply mic_gain, then use the AGC to limit the amplitude
// to 1.00
  vl_gain_and_agc(vl.mic_gain,vl.mic_agc_time,1);
  vl_level_meter(MIC_GAIN_LINE,1);
  vlstat=3;
  if(vl_traces_mode==0)
    {
    draw_vl_trace(VLFMT_REAL,1);
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==1)
    {
    draw_vl_trace(VLFMT_REAL,0);
    if(kill_all_flag) goto run_speech_x;
    }
  break;

  case 3:
// Filter the signal through audio filter 1.
  vl_filter(vl.af_filter_low,vl.af_filter_high,vl.af_filter_slope,1,2);
  vl_level_meter(AF_FILTER_LINE,1);
  vlstat=4;
  if(vl_traces_mode==1)
    {
    draw_vl_trace(VLFMT_REAL,1);
    if(kill_all_flag) goto run_speech_x;
    }
  break;

  case 4:
// Apply af_gain, then use the AGC to limit the amplitude
// to 1.00
  vl_gain_and_agc(vl.af_gain,vl.af_agc_time,1);
  vl_level_meter(AF_GAIN_LINE,1);
  if(vl_traces_mode==0)
    {
    draw_vl_trace(VLFMT_REAL,2);
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==2)
    {
    draw_vl_trace(VLFMT_REAL,0);
    if(kill_all_flag) goto run_speech_x;
    }
  vlstat=5;
  break;

  case 5:
// Filter the signal through RF filter 1.
  vl_filter(vl.rf_filter1_low,vl.rf_filter1_high,vl.rf_filter1_slope,2,2);
  vl_level_meter(RF_FILTER1_LINE,2);
  vlstat=6;
  break;

  case 6:
// Apply rf_gain1, then use the AGC to limit the amplitude.
// to 1.00
  vl_gain_and_agc(vl.rf_gain1,vl.rf_agc_time1,2);
  vl_level_meter(RF_GAIN1_LINE,2);
  vlstat=7;
  if(vl_traces_mode==0)
    {
    draw_vl_trace(VLFMT_CMPLX_AMPL,3);
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==1)
    {
    draw_vl_trace(VLFMT_CMPLX_AMPL,2);
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==2)
    {
    draw_vl_trace(VLFMT_CMPLX_AMPL,1);
    if(kill_all_flag) goto run_speech_x;
    }
  break;

  case 7:
// Filter the signal through RF filter 2.
  vl_filter(vl.rf_filter2_low,vl.rf_filter2_high,vl.rf_filter2_slope,2,1);
  vl_level_meter(RF_FILTER2_LINE,2);
  vlstat=8;
  break;

  case 8:
// Apply rf_gain2, then use the AGC to limit the amplitude.
// to 1.00
  vl_gain_and_agc(vl.rf_gain2,vl.rf_agc_time2,2);
  vl_level_meter(RF_GAIN2_LINE,2);
  vlstat=9;
  if(vl_traces_mode==2)
    {
    draw_vl_trace(VLFMT_CMPLX_AMPL,2);
    if(kill_all_flag) goto run_speech_x;
    }
  break;

  case 9:
// Filter the signal through RF filter 3.
  vl_filter(vl.rf_filter3_low,vl.rf_filter3_high,vl.rf_filter3_slope,2,1);
  vl_level_meter(RF_FILTER3_LINE,2);
  vlstat=10;
  break;

  case 10:
// Apply ALC to set the amplitude to 1.00
  vl_gain_and_agc(vl.alc_gain,vl.alc_time,2);
  vl_level_meter(ALC_LINE,2);
  vlstat=101;
  if(vl_traces_mode==0)
    {
    draw_vl_trace(VLFMT_REAL,4);
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==1)
    {
    draw_vl_trace(VLFMT_REAL,3);
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==2)
    {
    draw_vl_trace(VLFMT_CMPLX_AMPL,3);  
    if(kill_all_flag) goto run_speech_x;
    }
  if(vl_traces_mode==2)
    {
    draw_vl_trace(VLFMT_REAL,4);
    if(kill_all_flag) goto run_speech_x;
    }
  break;

  case 102:
// ************************************************
// The Phonetics is now sent to the output device.
  if(waitsamps <= vl_size/8)
    {
    vlstat=103;
    waitsamps+=tx.input_speed*vl.wait_time+wrsamps;
    if(guess_flag != 0 && fixed_phonetics_flag == 0)vl_missed_guesses++;
    guess_flag=current_phonetics;
    }
  break;

  case 103:
  if(waitsamps <= 0)vlstat=0;
  break;
  }
// Make sure the circular buffer raw_output[] is always filled with
// random_noise.
while(raw_output_px != raw_output_pa)
  {
  raw_output[raw_output_pa]=noisegen();
  raw_output_pa=(raw_output_pa+1)&mask;
  }
if(vlstat == 101)
  {
// Now the current Phonetics will be sent to the output.
// Allow the user his time to hit the right key on the keyboard.
// This block is placed outside the switch to ensure it is always
// operating on a raw_output[] which is already filled with noise.
  waitsamps=vl_size+wrsamps/2;
  vlstat=102;
  ia=raw_output_px+wrsamps/2;
  for(i=0; i<vl_size; i++)
    {
    ia=(ia+1)&mask;
    raw_output[ia]+=VOICELAB_AMPLITUDE*vl_buffer[2*i  ];
    }
  }
// Filter a block of data from raw_output to filtered output
// if there is free space enough.
i=(filtered_output_px-filtered_output_pa+mask)&mask;
if(i >= wrsamps)
  {
// Copy 2*wrsamps (=rx_fftsize) points to rxbuf while
// multiplying with the sin squared window.
  for(i=0; i<wrsamps; i++)
    {
    rxbuf[2*i]=raw_output[raw_output_px]*rx_fft_win[i];
    rxbuf[2*i+1]=0;
    raw_output_px=(raw_output_px+1)&mask;
    }
  ia=raw_output_px;
  k=wrsamps;
  for(i=wrsamps; i<(wrsamps<<1); i++)
    {
    rxbuf[2*i]=raw_output[ia]*rx_fft_win[k];
    rxbuf[2*i+1]=0;
    ia=(ia+1)&mask;
    k--;
    }
  fftforward(rx_fftsize, rx_fftn, rxbuf, rx_fft_table, rx_fft_permute);
// Clear points outside the selected passband
  j=rx_fftsize*vl.rx_filter_low/tx.input_speed;
  k=rx_fftsize*vl.rx_filter_high/tx.input_speed;
  for(i=0; i<2*j; i++)rxbuf[i]=0;
  for(i=2*k; i<2*rx_fftsize; i++)rxbuf[i]=0;
// Find out how many bins there are for 1 kHz.
// set a factor that will give the desired slope.
  t1=(float)(rx_fftsize*2000)/tx.input_speed;
  t2=pow(10.,vl.rx_filter_slope*0.05);
  t1=pow(t2,1/t1);
  t2=t1;
  i=(j+k)/2;
  i--;
  j=i+2;
  while(i<=k)
    {
    rxbuf[2*i]*=t2;
    rxbuf[2*i+1]*=t2;
    rxbuf[2*j]/=t2;
    rxbuf[2*j+1]/=t2;
    t2*=t1;
    i++;
    j--;
    }
  fftback(rx_fftsize, rx_fftn, rxbuf, rx_fft_table, rx_fft_permute);
  for(i=wrsamps>>1; i<wrsamps; i++)
    {
    filtered_output[filtered_output_pa]=rxbuf[2*i]/(rx_fft_win[i]*rx_fftsize);
    filtered_output_pa=(filtered_output_pa+1)&mask;
    }
  k=wrsamps;
  for(i=wrsamps; i<3*(wrsamps>>1); i++)
    {
    filtered_output[filtered_output_pa]=rxbuf[2*i]/(rx_fft_win[k]*rx_fftsize);
    filtered_output_pa=(filtered_output_pa+1)&mask;
    k--;
    }
  }  
if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
  {
  lirerr(28844);
  goto run_speech_x;
  }
if(ad_info.bytes > ad_info.fragsize*ad_info.fragstotal-tx.input_speed*0.05)
  {
  t1=0;
  if(rx_daout_channels == 1)
    {
    for(i=0; i<wrsamps; i++)
      {
      wrbuf[i]=filtered_output[filtered_output_px];
      if(fabs(filtered_output[filtered_output_px]) >t1)
                                  t1=fabs(filtered_output[filtered_output_px]);
      filtered_output_px=(filtered_output_px+1)&mask;
      }
    }
  else
    {
    for(i=0; i<wrsamps/2; i++)
      {
      wrbuf[2*i]=filtered_output[filtered_output_px];
      wrbuf[2*i+1]=filtered_output[filtered_output_px];
      if(fabs(filtered_output[filtered_output_px]) >t1)
                                  t1=fabs(filtered_output[filtered_output_px]);
      filtered_output_px=(filtered_output_px+1)&mask;
      }
    }
  t1=100*t1/0x7fff;
  if(t1 > max_output)max_output=t1;
  sprintf(s,"D/A:%4.1f%%    mode=%d ",max_output,vl_traces_mode);  
  lir_text(LEVEL_METER_X+12,3,s);
  if(waitsamps > 0) waitsamps-=wrsamps/2;
  write(rx_audio_out,wrbuf,rx_daout_block);
  }

test_keyboard();
if(kill_all_flag) goto run_speech_x; 
if(lir_inkey != 0)
  {
  lir_inkey=toupper(lir_inkey);
  switch (lir_inkey)
    {
    case '^':
    vl_traces_mode ++;
    if(vl_traces_mode > 2)vl_traces_mode=0;
    break;

    case '"':
    vl_signal_type ++;
    if(vl_signal_type > 4)vl_signal_type=0;
    break;
    
    case '#':
    clear_lines(screen_last_line-1,screen_last_line);
    close_rx_sndout();    
    if(kill_all_flag) goto run_speech_x;
    lir_text(0,screen_last_line,"Press key for fixed Phonetics. '*' to skip");
get_fixed:;
    await_keyboard();
    if(kill_all_flag) goto run_speech_x;
    if(lir_inkey == '*')
      {
      fixed_phonetics_flag=0;
      }    
    else
      {
      lir_inkey=toupper(lir_inkey);
      if( (phonetics_flag[(int)(lir_inkey)]&2) == 0)goto get_fixed;
      fixed_phonetics_flag=lir_inkey;
      }
    clear_lines(screen_last_line,screen_last_line);
    open_rx_sndout();
    if(kill_all_flag) goto run_speech_x;
    goto sp_loop_init_b;

    case '*':
    close_rx_sndout();
    if(kill_all_flag)goto run_speech_x;
    
    case '?':
    clear_lines(screen_last_line-1,screen_last_line);
    close_rx_sndout();    
    if(kill_all_flag) goto run_speech_x;
    lir_text(0,screen_last_line,"Line to change >");
    i=lir_get_integer(16, screen_last_line, 4, 1, MAX_SPEECHPROC_PARMS);
    if(kill_all_flag) goto run_speech_x;
    clear_lines(screen_last_line,screen_last_line);
    change_speechproc_parm(i);
    if(kill_all_flag) goto run_speech_x;
    open_rx_sndout();
    if(kill_all_flag) goto run_speech_x;
    clear_lines(screen_last_line,screen_last_line);
    goto sp_loop_init_b;

    case '/':
    lir_inkey='G';
    process_current_lir_inkey();
    if(kill_all_flag) goto run_speech_x;
    break;

    case '+':
    vlt_step++;
    break;

    case '-':
    vlt_step--;
    if(vlt_step < 1)vlt_step=1;
    break;

    case '>':
    case ARROW_RIGHT_KEY:
    vlt_shift+=0.14*screen_width;
    break;

    case '<':
    case ARROW_LEFT_KEY:
    vlt_shift-=0.1*screen_width;
    break;

    case ',':
    close_rx_sndout();    
    if(kill_all_flag) goto run_speech_x;
    await_keyboard();
    if(kill_all_flag) goto run_speech_x;
    open_rx_sndout();
    if(kill_all_flag) goto run_speech_x;
    break;

    case '&':
    case ':':
    k=(lir_inkey - '&');
    close_rx_sndout();    
    if(kill_all_flag) goto run_speech_x;
    clear_lines(screen_last_line-1,screen_last_line);
    lir_text(0,screen_last_line,"File no (1 to 9) >");
    i=lir_get_integer(20, screen_last_line, 4, 1, 10);
    if(kill_all_flag) goto run_speech_x;
    clear_lines(screen_last_line,screen_last_line);
    sprintf(s,"voicepar%d.bin",i);
    if( k == 0)
      {
      file=fopen(s,"r");
      }
    else
      {  
      file=fopen(s,"w");
      }
    if(file==NULL)
      {
      lir_text(0,screen_last_line,"Could not open");
      lir_text(16,screen_last_line,s);
      lir_text(36,screen_last_line,"PRESS ANY KEY");
      await_keyboard();
      if(kill_all_flag) goto run_speech_x;
      }
    else
      {
      if(k == 0)
        {
        fread(&vl,sizeof(vl),1,file);
        }
      else
        {  
        fwrite(&vl,sizeof(vl),1,file);  
        }
      fclose(file);
      }
    clear_lines(screen_last_line,screen_last_line);
    open_rx_sndout();
    if(kill_all_flag) goto run_speech_x;
    goto sp_loop_init_a;
    
    default:
    if(fixed_phonetics_flag == 0)
      {
      settextcolor(7);
      if(guess_flag != 0)
        {
        if(lir_inkey == guess_flag)
          {
          vl_true_guesses++;
          clear_lines(screen_last_line-1,screen_last_line-1);
          }
        else
          {
          settextcolor(12);
          vl_false_guesses++;
          sprintf(s,"Phonetics:%c  Guess:%c", guess_flag, lir_inkey);
          lir_text(0,screen_last_line-1,s);
          settextcolor(7);
          }
        guess_flag=0;
        }
      show_guess_statistics();
      }
    else
      {
      if( (phonetics_flag[(int)(lir_inkey)]&2) != 0)
                               fixed_phonetics_flag=lir_inkey;
      }
    settextcolor(7);
    break;      
    }
  }
goto sp_loop;
run_speech_x:;
memcheck(2999,voicelabmem,(int*)&voicelab_handle);
free(voicelab_handle);    
free(txmem_handle);    
}

void show_phonetics_info(int inum, char colour)
{
char s[80];
settextcolor(colour);
if( (phonetics_flag[inum]&2)!=0)
  {
  sprintf(s,"%c %.2fdB %.2fs     ",inum,phonetics_ampl[inum],
    (float)(phonetics_pointer[inum+1]-phonetics_pointer[inum])/tx.input_speed);
  s[18]=0;
  }
else
  {
  sprintf(s,"%c - - - - ",inum);  
  }
lir_text(phonetics_col[inum],phonetics_line[inum],s);
}

void get_phonetics(void)
{
char s[80];
int inum;
int line,col;
int ia,ib,ic,ix,ixinc,iadec;
int i, j, k, nstep, pa, pb, pc, im, px, mask;
int imax, imin;
int imax1, imin1;
int satflag, recsamps;
char *rdbuf;
short int *rdbuf_short_int;
int *rdbuf_int;
audio_buf_info ad_info;
float t1, t2, t3, amax;
float *block_maxamp;
float *phonbuf, *rdbuf_float;
float voicelab_blocktime;
int phonbuf_size;
// This routine is for recording phonetics.
// Read A/D input and store waveform in a circular recording buffer.
// When recording is stopped, save up to two seconds of data
// in phonetics[] 
// Make the read block 1/32 seconds and allocate 8 seconds
// for the circular buffer
#define VOICELAB_BLOCKS_PER_SECOND 32
#define VOICELAB_TOTAL_TIME 8
#define VOICELAB_BLOCKS (VOICELAB_BLOCKS_PER_SECOND*VOICELAB_TOTAL_TIME)
tx_read_bytes=tx.input_speed*tx.input_channels*
                                   tx.input_bytes/VOICELAB_BLOCKS_PER_SECOND;
make_power_of_two(&tx_read_bytes);
voicelab_blocktime=(float)(tx_read_bytes)/
                           (tx.input_speed*tx.input_channels*tx.input_bytes);
block_maxamp = malloc(VOICELAB_BLOCKS*sizeof(float));
rdbuf = malloc(tx_read_bytes*VOICELAB_BLOCKS);
phonbuf_size=2.5*tx.input_speed;
phonbuf = malloc(phonbuf_size*sizeof(float));
if( block_maxamp == NULL || rdbuf == NULL || phonbuf == NULL)
  {
  lirerr(1044);
  goto get_phonetics_x;
  }
rdbuf_int=(void*)rdbuf;
rdbuf_short_int=(void*)rdbuf;
rdbuf_float=(void*)rdbuf;
mask=tx_read_bytes*VOICELAB_BLOCKS-1;
clear_screen();
lir_text(0,0,"Press keyboard key to activate character.");
lir_text(0,1,"Space bar to record.  Enter to stop and/or play.");
lir_text(0,2,"'End' or 'Home' to exit");
line=4;
col=0;
for(inum=0; inum<256; inum++)
  {
  if(phonetics_flag[inum]!=0)
    {
    phonetics_line[inum]=line;
    phonetics_col[inum]=col;
    show_phonetics_info(inum,7);
    col+=PHONETICS_COLINC;
    if((col + PHONETICS_COLINC)*text_width > screen_last_xpixel)
      {
      col=0;
      line++;
      }
    }
  }      
lir_inkey=0;
while(lir_inkey != HOME_KEY && lir_inkey != END_KEY)
  {
  if(phonetics_flag[lir_inkey] != 0)
    {
    inum=lir_inkey;
playrec_1:;
    show_phonetics_info(inum,15);
playrec:;
    await_keyboard();
    if(kill_all_flag) goto get_phonetics_x;
    lir_inkey=toupper(lir_inkey);
    if(lir_inkey == ' ')
      {
      for(i=0; i<VOICELAB_BLOCKS; i++)block_maxamp[i]=0;
      show_phonetics_info(inum,12);
      open_tx_input();
      if(kill_all_flag) goto get_phonetics_x;
      pa=0;
      nstep=tx.input_channels*tx.input_bytes;
      px=mask+1;  
      pb=nstep;
      read(tx_audio_in,rdbuf,tx_read_bytes); 
// Read and skip data for 0.25 seconds. 
// Some soundcards give a transcient due to the DC component
// that will decay slowly.     
      i=8;
      ad_info.bytes=-1;
      while(i > 0)
        {
        if(ioctl(tx_audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
          {
          lirerr(231028);
          goto get_phonetics_x;
          }
        if(ad_info.bytes==-1)
          {
          lirerr(231029);
          goto get_phonetics_x;
          }
        if(ad_info.bytes > tx_read_bytes)
          {
          read(tx_audio_in,rdbuf,tx_read_bytes); 
          i--;
          }
        }  
      satflag=0;
      settextcolor(7);
      imax=0x80000000;
      imin=0x7fffffff;
      while(lir_inkey != 10)
        {
        test_keyboard();
        if(kill_all_flag) goto get_phonetics_x;
        if(lir_inkey != 0)
          {
          process_current_lir_inkey();
          if(kill_all_flag) goto get_phonetics_x;
          }
        if(ioctl(tx_audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
          {
          lirerr(231028);
          goto get_phonetics_x;
          }
        if(ad_info.bytes > tx_read_bytes)
          {
          if(pa == px)px=px+tx_read_bytes;
          px&=mask;
          read(tx_audio_in,&rdbuf[pa],tx_read_bytes); 
          pa=(pa+tx_read_bytes)&mask;
          if(ioctl(tx_audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
            {
            lirerr(231028);
            goto get_phonetics_x;
            }
          imax1=0x80000000;
          imin1=0x7fffffff;
          t1=0;
          t2=0;
          pc=pb;
          im=pb/tx_read_bytes;
          if(tx.input_bytes==2)
            {
            while(pb != pa)
              {
              i=rdbuf_short_int[pb>>1];
              if(i>imax1)imax1=i;
              if(i<imin1)imin1=i;
              t1+=i;
              pb=(pb+nstep)&mask;
              }
            t1=nstep*t1/tx_read_bytes; 
            while(pc != pa)
              {
              t3=rdbuf_short_int[pc>>1]-t1;
              t3/=0x7fff;
              t2+=t3*t3;  
              pc=(pc+nstep)&mask;
              }
            block_maxamp[im]=nstep*t2/tx_read_bytes;
            if(imax == 0x7fff || imin == 0x8000)satflag=1;
            if(imax1>imax)imax=imax1;
            if(imin1<imin)imin=imin1;
            t1=((float)(imax)-(float)(imin))/0x10000;
            t2=((float)(imax1)-(float)(imin1))/0x10000;
            }
          else
            {  
            while(pb != pa)
              {
              i=rdbuf_int[pb>>2];
              if(i>imax1)imax1=i;
              if(i<imin1)imin1=i;
              t1+=i;
              pb=(pb+nstep)&mask;
              }
            t1=nstep*t1/tx_read_bytes; 
            while(pc != pa)
              {
              t3=rdbuf_short_int[pc>>1]-t1;
              t3/=0x7fff0000;
              t2+=t3*t3;  
              pc=(pc+nstep)&mask;
              }
            block_maxamp[im]=nstep*t2/tx_read_bytes;
            if(imax >=(int) 0x7fff0000 || imin <= (int)0x8000ffff)satflag=1;
            if(imax1>imax)imax=imax1;
            if(imin1<imin)imin=imin1;
            t1=((float)(imax)-(float)(imin))/0x7fffffff;
            t2=((float)(imax1)-(float)(imin1))/0x7fffffff;
            }
          if(ad_info.bytes < tx_read_bytes)
            {
            sprintf(s,"%5d   %5d  ",((pa-px+mask)&mask)/tx_read_bytes,
                                     pa/tx_read_bytes);
            lir_text(0,line+3,s); 
            if(t1>0 && t2>0)
              {
              sprintf(s,"%.2fdB   %.2fdB  ",-20*log10(t1),-20*log10(t2));
              lir_text(0,line+5,s); 
              }
            if(satflag != 0)
              {
              settextcolor(12);
              for(i=0; i<4; i+=2)lir_text(0,line+4+i,"SATURATED");
              settextcolor(7);
              }
            }
          }
        }
      close_tx_input();
      if(kill_all_flag) goto get_phonetics_x;
// Now we have a recording in the circular buffer.
// block_maxamp[] gives the AC power over a period of
// 1/VOICELAB_BLOCKS_PER_SECOND seconds.
// Find where the power is maximum.
      t1=0;
      im=0;
      for(i=0; i<VOICELAB_BLOCKS; i++)
        {
        if(t1 < block_maxamp[i])
          {
          t1=block_maxamp[i];
          im=i;
          }
        }
// The data is in a circular buffer. px and pa point to the beginning
// and to the end of the recording. Convert to pointers in block_maxamp[]
// and move the pointers so the time spanned is 2 seconds or less while
// skipping the part of the recording with lowest signal level.      
      ia=pa/tx_read_bytes;
      ix=px/tx_read_bytes;
      ixinc=0;
      iadec=0;
      j=VOICELAB_BLOCKS-1;
      ia=(ia+j)&j;
// First skip blocks with less than -20dB signal level from both ends.
      while(block_maxamp[ix] < 0.01*t1)
        {
        ix=(ix+1)&j;
        ixinc++;
        }
      while(block_maxamp[ia] < 0.01*t1)
        {
        ia=(ia+j)&j;
        iadec++;
        }
      i=(ia-ix+j)&j;        
      t2=(float)(i)*voicelab_blocktime;
      clear_lines(line+1,line+9);
      if(t2 > 2)
        {
        for(j=1; j<6; j++)
          {
          settextcolor(12);
          lir_text(0,line+j,"ERROR:");
          settextcolor(15);
          lir_text(9,line+j,"Signal duration > 2 sec. (-20dB)");
          }
        }
      else
        {
// The Phonetics lasts for less than 2 seconds.
// include blocks down to -30 dB up to a total time of 2 seconds.
        sprintf(s,"Valid range (-20dB) %d to %d.   Max %d",ix,ia,im);
        lir_text(0,line+2,s);      
        t1=(float)(i+1)*voicelab_blocktime;
        while(t1 <= 2 && (ixinc > 0 || iadec > 0))
          {
          ib=0;
          ic=0;
          if(iadec > 0)
            {
            ib=(ia+1)&j;
            t2=block_maxamp[ib];
            }
          else
            {
            t2=-1;
            }  
          if(ixinc > 0)
            {
            ic=(ix+j)&j;
            t3=block_maxamp[ic];
            }
          else
            {
            t3=-1;
            }  
          if(t2 > t3)
            {
            ia=ib;
            iadec--;
            }
          else
            {
            ix=ic;
            ixinc--;
            }
          i=(ia-ix+j)&j;        
          t1=(float)(i+1)*voicelab_blocktime;
          }
        sprintf(s,"valid range (-30dB) %d to %d.   Max %d",ix,ia,im);
        lir_text(0,line+3,s);      
// Convert to floating point and move to phonbuf[].
// Include one block outside valid data.
// They will be used to remove very low frequencies, but
// they will not be saved.
        if(ixinc != 0)ix=(ix+j)&j;
        if(iadec != 0)ia=(ia+1)&j;
        recsamps=0;
        pa=ia*tx_read_bytes;
        px=ix*tx_read_bytes;
        if(tx.input_bytes==2)
          {
          while(px != pa)
            {
            phonbuf[recsamps]=(float)(rdbuf_short_int[px>>1])/0x7fff;
            px=(px+nstep)&mask;
            recsamps++;
            if(recsamps >= phonbuf_size) 
              {
              lirerr(232222);
              goto get_phonetics_x;
              }
            }
          }
        else
          {
          while(px != pa)
            {
            phonbuf[recsamps]=(float)(rdbuf_int[px>>2])/0x7fffff00;
            px=(px+nstep)&mask;
            recsamps++;
            if(recsamps >= phonbuf_size)
              {
              lirerr(232222);
              goto get_phonetics_x;
              }
            }
          }
// High-pass filter phonbuf to get rid of any DC component
// and possible decay of switch transcients at the start
// of the recording.        
// Store the filtered data in rdbuf_float 
// (we will not need the raw data any more).        
        j=tx_read_bytes/nstep;
        t1=0;
        for(i=0; i<2*j; i++)
          {
          t1+=phonbuf[i];
          }
        for(i=j; i<recsamps-j; i++)
          {
          rdbuf_float[i-j]=phonbuf[i]-t1/(2*j);
          t1-=phonbuf[i-j];
          t1+=phonbuf[i+j];
          }
        recsamps-=j;
// We may have included too many samples in the waveform.
// Put the integrated power into phonbuf[].
        t1=0;
        t2=0;   
        for(i=0; i<recsamps; i++)
          {
          t3=rdbuf_float[i]*rdbuf_float[i];
          t1+=t3;
          if(t2<t3)t2=t3;
          phonbuf[i]=t1;
          }
// Point ia and ib to the first point above -30dB below maximum
// for which the total loss of spoken energy is below 1%.
        j=0.02*tx.input_speed;
        ib=recsamps-1;
        while( (1000*(t1-phonbuf[ib]) < t1 || 
                1000*rdbuf_float[ib]*rdbuf_float[ib] < t2) && ib > j)ib--;
        ia=0;
        while( (1000*phonbuf[ia] < t1 || 
                1000*rdbuf_float[ia]*rdbuf_float[ia] < t2) && ia < ib-j)ia++;
        t1=(float)(ib-ia)/tx.input_speed;
        sprintf(s,"Time %f seconds",t1);
        lir_text(0,line+4,s);      
// The DC level may still be inaccurate.
// Subtract the DC component by subtracting a straight line
// that makes the average value of the first and last 20 milliseconds zero.
        t1=0;
        for(i=0; i<j; i++)t1+=rdbuf_float[i+ia];
        t1/=j;
        t2=0;
        for(i=0; i<j; i++)t2+=rdbuf_float[ib-1-i];
        t2/=j;
        k=ib-ia;
        t1/=k;
        t2/=k;
        recsamps=0;
        for(i=ia; i<ib; i++)
          {
          phonbuf[recsamps]=rdbuf_float[i]-t1*(recsamps-k)-t2*recsamps;
          recsamps++;
          }  
// Attenuate switch transcients by applying a cosine squared window
// at the beginning and at the end of the recording.
        t1=PI_L/j;
        t2=t1;
        ia=0;
        ib=recsamps-1;
        while(t2 < 0.5*PI_L)
          {
          t3=sin(t2);
          t3=t3*t3;
          phonbuf[ia]*=t3;
          phonbuf[ib]*=t3;
          ia++;
          ib--;
          t2+=t1;
          }
// Point ia and ib to the beginning and end of any previous recording
// for this character.
        amax=0;
        for(i=0; i<recsamps; i++)
          {
          if(fabs(phonbuf[i]) > amax)amax=fabs(phonbuf[i]);
          }
        amax=20*log10(amax);  
        ia=phonetics_pointer[inum];
        ib=phonetics_pointer[inum+1];        
        ic=phonetics_pointer[255];
        k=ib-ia;
        j=PHONETICS_SIZE-ic+k;
        if(j<recsamps)
          {
          lirerr(232233);
          goto get_phonetics_x;
          }
        sprintf(s,"Size %d samples. %fdB  Remaining storage %d",
                                                           recsamps,amax,j);
        lir_text(5,line+6,s);
        sprintf(s,"Press 'S' to store recording for [%c]  ",inum);
        lir_text(5,line+7,s);
        lir_text(5,line+8,"Any other key to skip");
        await_keyboard();
        if(kill_all_flag) goto get_phonetics_x;
        if(toupper(lir_inkey) == 'S')
          {
// Move recordings for higher valued characters to create a free space
// of recsamps points.            
          pb=ia+recsamps;
          k=pb-ib;
          if(k >= 0)
            {
            for(i=ic; i>=ib; i--)
              {
              phonetics[i+k]=phonetics[i];
              }
            }
          else
            {
            for(i=ib; i<=ic; i++)
              {
              phonetics[i+k]=phonetics[i];
              }
            }
          for(i=0; i<recsamps; i++)
            {
            phonetics[ia+i]=phonbuf[i];
            }
          j=inum+1;
          while(j<256)
            {
            phonetics_pointer[j]+=k;
            j++;
            }
          phonetics_ampl[inum]=amax;
          phonetics_flag[inum]|=2;          
          clear_lines(line+1,line+9);
          goto playrec_1;      
          }
        else
          {  
          clear_lines(line+1,line+9);
          goto playrec;      
          }
        }
      settextcolor(15);
      lir_text(10,line+9,"PRESS ANY KEY");
      await_keyboard();
      if(kill_all_flag) goto get_phonetics_x;
      clear_lines(line+1,line+9);
      goto playrec;      
      }
    if(lir_inkey == 10)
      {
      if((phonetics_flag[inum] & 2) == 0)goto playrec_1;
      show_phonetics_info(inum,14);
      ia=phonetics_pointer[inum];
      ib=phonetics_pointer[inum+1]-1;
// The data in phonetics is in floating point format at a sampling speed
// of tx.input_speed.
// Convert to short int in rdbuf_short_int and set the rx 
// output for the same speed.
      recsamps=0;
      while(ia != ib)
        {
        rdbuf_short_int[recsamps]=phonetics[ia]*0x7fff;
        ia++;
        recsamps++;
        }
      open_rx_sndout();
      if(kill_all_flag) goto get_phonetics_x;
      i=0;  
      k=recsamps;
      while(i<rx_daout_block)
        {
        rdbuf_short_int[recsamps]=0;
        recsamps++;
        i++; 
        }
      write(rx_audio_out, rdbuf_short_int, rx_daout_block);
      if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
        {
        lirerr(238844);
        goto get_phonetics_x;
        }
      i=rx_daout_block/2;
      while(i <= recsamps)
        {
        if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
          {
          lirerr(238844);
          goto get_phonetics_x;
          }
        if(ad_info.bytes > rx_daout_block)
          {
          write(rx_audio_out,&rdbuf_short_int[i],rx_daout_block);
          i+=rx_daout_block/2;
          }
        else
          {
          lir_sched_yield();
          }  
        }  
      i=0;
      for(k=0; k<rx_daout_block/2; k++)rdbuf_short_int[k]=0;
      k=0;
      while(i<4)
        {
        if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
          {
          lirerr(238844);
          goto get_phonetics_x;
          }
        if(ad_info.bytes > rx_daout_block)
          {
          write(rx_audio_out,&rdbuf_short_int[k],rx_daout_block);
          i++;
          }
        else
          {
          lir_sched_yield();
          }  
        }  
      while(ad_info.fragstotal-ad_info.fragments > 4)
        {
        if(ioctl(rx_audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
          {
          lirerr(248844);
          goto get_phonetics_x;
          }
        }
      close_rx_sndout();
      if(kill_all_flag) goto get_phonetics_x;
      goto playrec_1;      
      }
    show_phonetics_info(inum,7);
    }
  else
    {
    await_keyboard();
    lir_inkey=toupper(lir_inkey);
    }
  } 
get_phonetics_x:;
free(rdbuf);
}

int diskread_phonetics(FILE *labfile)
{
unsigned int ia, retcod;
retcod=0;
if(fread(phonetics_flag,sizeof(int),256,labfile) != 256)goto errexit;
if(fread(phonetics_pointer,sizeof(int),256,labfile) != 256)goto errexit;
if(fread(phonetics_ampl,sizeof(float),256,labfile) != 256)goto errexit;
ia=phonetics_pointer[255];
if(fread(phonetics,sizeof(float),ia,labfile) != ia)
  {
errexit:;
  retcod=-1;
  }   
fclose(labfile);  
return retcod;
}

void speech_proc_lab(void)
{
char s[80];
int i,j;
unsigned int ia;
FILE *labfile;
char labfilenam[]={"voicelab.bin"};
phonetics = malloc(PHONETICS_SIZE*sizeof(float));
if(phonetics == NULL)
  {
  lirerr(1044);
  return;
  }
labfile=fopen(labfilenam,"rb");
if(labfile != NULL)
  {
  i=diskread_phonetics(labfile);
  if(i != 0)goto errfile;
  }
else
  {  
errfile:;  
  for(i=0; i<256; i++)
    {
    phonetics_flag[i]=0;
    phonetics_pointer[i]=0;
    phonetics_ampl[i]=-999;
    }
  for(i='0'; i<='9'; i++)phonetics_flag[i]=1;
  for(i='A'; i<='Z'; i++)phonetics_flag[i]=1;
  } 
lir_inkey=0;   
while (lir_inkey != 'X')
  {
  clear_screen();
  settextcolor(15);
  lir_text(25,0,"Speech Processing Lab");
  settextcolor(7);
  lir_text(5,3,"A=Record or edit Phonetics"); 
  lir_text(5,4,"B=Evaluate speech processing"); 
  lir_text(5,5,"R=Read Phonetics from disk");
  lir_text(5,6,"S=Save Phonetics on  disk");
  lir_text(5,7,"X=Return");
  await_processed_keyboard();
  if(kill_all_flag) return;
  switch(lir_inkey)
    {
    case 'A':
    if(ui.max_output_bytes < 2)lirerr(258799);
    get_phonetics();
    if(kill_all_flag) return;
    break;

    case 'B':
    if(ui.max_output_bytes < 2)lirerr(268799);
    j=0;
    for(i=0; i<256; i++)
      {
      if( (phonetics_flag[i]&2) != 0)j++;
      }
    if(j < 10)
      {
      clear_screen();
      lir_text(5,5,"ERROR:  Store Phonetics for at least 10 characters");
      lir_text(5,6,"before evaluating speech processing.  PRESS ANY KEY");
      await_keyboard();
      }
    else
      {    
      run_speech_processing();
      if(kill_all_flag) return;
      }
    break;

    case 'R':
    labfile=fopen(labfilenam,"rb");
    if(labfile == NULL)
      {
      clear_screen();
      sprintf(s,"Could not open %s",labfilenam);
      lir_text(5,5,s);
      lir_text(5,7,"PRESS ANY KEY");
      await_keyboard();
      }
    else
      {     
      i=diskread_phonetics(labfile);
      if(i != 0)
        {
        clear_screen();
        sprintf(s,"Error in %s",labfilenam);
        lir_text(5,5,s);
        lir_text(5,7,"PRESS ANY KEY");
        await_keyboard();
        goto errfile;
        }
      }
    break;

    case 'S':
    labfile=fopen(labfilenam,"wb");
    if(labfile == NULL)lirerr(278799);
    if(fwrite(phonetics_flag,sizeof(int),256,labfile) != 256)
                                                       lirerr(288788); 
    if(fwrite(phonetics_pointer,sizeof(int),256,labfile) != 256)
                                                       lirerr(308787); 
    if(fwrite(phonetics_ampl,sizeof(float),256,labfile) != 256)
                                                       lirerr(318787); 
    ia=phonetics_pointer[255];
    if(fwrite(phonetics,sizeof(float),ia,labfile) != ia)lirerr(328786); 
    fclose(labfile);  
    break;
    }
  }
free(phonetics);
}
#endif
