
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "windef.h"
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "conf.h"
#include "txdef.h"
#include "vernr.h"
#include "screendef.h"
#include "sigdef.h"
#include "thrdef.h"
#include "sdrdef.h"

// allocate 16 megabytes for phonetics (float format).
#define PHONETICS_SIZE (1024*4096)
#define VOICELAB_AMPLITUDE (0.2*0x7fff)
#define LEVEL_METER_X 65

extern void win_query_warning(int *line);

VOID CALLBACK tx_daout(HWAVEOUT hwi, UINT iMsg, DWORD dwInst, 
                         DWORD dwParam1, DWORD dwParam2)
{
DWORD i;
i=(DWORD)hwi;
i=dwInst;
i=dwParam2;
i=iMsg;
i=dwParam1;
if(iMsg == WIM_DATA)
  {
  }
else
  {
  if(iMsg == WIM_CLOSE)
    {
    i=(DWORD)hwi;
    i=dwInst;
    i=dwParam2;
    tx_audio_out=-1;
    }
  else
    {
    if(iMsg == WIM_OPEN)
      {
      tx_audio_out=0;
      }
    }  
  }
}                         

void open_tx_output(void)
{
int errcod;
WAVEFORMATEXTENSIBLE wfmt;
MMRESULT mmrs;
int i, k;
if(tx_audio_out != -1)return;
wfmt.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
wfmt.Format.nSamplesPerSec=tx.output_speed;
fprintf(dmp,"\nspeed %d",tx.output_speed);
if(tx.output_bytes == 2)
  {
  wfmt.Samples.wValidBitsPerSample=16;
  wfmt.Format.wBitsPerSample=16;
  }
else
  {
  wfmt.Samples.wValidBitsPerSample=24;
  wfmt.Format.wBitsPerSample=32;
  }
wfmt.SubFormat = KSDATAFORMAT_SUBTYPE_PCM;
wfmt.dwChannelMask = 7;
wfmt.Format.nChannels=tx.output_channels;
fprintf(dmp,"\nchannels %d",tx.output_channels);
wfmt.Format.nBlockAlign=(unsigned int)((wfmt.Format.nChannels *
                                         wfmt.Format.wBitsPerSample) / 8.0);
wfmt.Format.nAvgBytesPerSec=wfmt.Format.nBlockAlign*wfmt.Format.nSamplesPerSec;
wfmt.Format.cbSize=sizeof(WAVEFORMATEXTENSIBLE);;
fprintf(dmp,"\ndev %d",tx.da_device_no);
mmrs=waveOutOpen(&hwav_txdaout, tx.da_device_no, 
              (WAVEFORMATEX*)&wfmt, (DWORD)tx_daout, 0, CALLBACK_FUNCTION);
if(mmrs != MMSYSERR_NOERROR)
  {
  lirerr(1234);
  return;
  }
errcod=1237;
tx_waveout_buf=malloc(NO_OF_TX_WAVEOUT*tx_daout_block+wfmt.Format.nBlockAlign);
if(tx_waveout_buf==NULL)goto wave_close;
tx_wave_outhdr=malloc(NO_OF_TX_WAVEOUT*sizeof(WAVEHDR));
if(tx_wave_outhdr==NULL)goto free_waveout;
tx_da_wrbuf=malloc(tx_daout_block);
if(tx_da_wrbuf==NULL)goto free_outhdr;
txdaout_newbuf_ptr=0;
for(i=0; i<NO_OF_TX_WAVEOUT; i++)
  {
  k=(int)(tx_waveout_buf+i*tx_daout_block+wfmt.Format.nBlockAlign);
  k&=(-1-wfmt.Format.nBlockAlign);
  tx_wave_outhdr[i].lpData=(char*)(k);
  tx_wave_outhdr[i].dwBufferLength=tx_daout_block;
  tx_wave_outhdr[i].dwBytesRecorded=0;
  tx_wave_outhdr[i].dwUser=1; 
  tx_wave_outhdr[i].dwFlags=0;
  tx_wave_outhdr[i].dwLoops=0;
  tx_wave_outhdr[i].lpNext=NULL;
  tx_wave_outhdr[i].reserved=0;
  mmrs=waveOutPrepareHeader(hwav_txdaout, &tx_wave_outhdr[i], sizeof(WAVEHDR));
  txdaout_newbuf[i]=&tx_wave_outhdr[i];
  if(mmrs != MMSYSERR_NOERROR)goto errfree;
  }
tx_da_totbytes=NO_OF_TX_WAVEOUT*tx_daout_block;
tx_audio_out=0;
return;
errfree:;
free(tx_da_wrbuf);
free_outhdr:;
free(tx_wave_outhdr);
free_waveout:;
free(tx_waveout_buf);
wave_close:;
waveOutClose(hwav_txdaout);
tx_audio_out=-1;
lirerr(errcod);
}


/*
void close_tx_output(void)
{
if(close(tx_audio_out)==-1)lirerr(231037);  
tx_audio_out=-1;
}
*/



void open_tx_input(void)
{
/*
int i, j, frag;
float t1;
xz("open tx");
make_devname(tx.ad_device_no);
tx_audio_in=open( dev_name, O_RDONLY, 0);
if(tx_audio_in == -1)
  {
  lirerr(1260);
  return;
  }
// We want the fragment size to match tx_read_bytes.
frag=0;
i=tx_read_bytes>>1;
while(i>0)
  {
  i>>=1;
  frag++;
  }
if(frag < 4)frag=4;
frag=frag|0x7fff0000;
if(ioctl(tx_audio_in, SNDCTL_DSP_SETFRAGMENT, &frag) == -1)
  {
  lirerr(1262);
  return;
  }
i=AFMT_S16_LE;
#ifdef AFMT_S32_LE
if(tx.input_bytes==4)i=AFMT_S32_LE;
#else
tx.input_bytes=2;
#endif
j=i;
if(ioctl(tx_audio_in, SNDCTL_DSP_SETFMT, &i) == -1)
  {
  lirerr(231013);
  return;
  }
if(i!=j)
  {
  lirerr(231014);
  return;
  }
// Set number of channels
j=tx.input_channels;
i=j;
if(ioctl(tx_audio_in, SNDCTL_DSP_CHANNELS, &i) == -1)
  {
  lirerr(231009);  
  return;
  }
if(i != j)
  {
  lirerr(231010);
  return;
  }
i=tx.input_speed;
if(ioctl(tx_audio_in, SNDCTL_DSP_SPEED, &i)==-1)
  {
  lirerr(231011);
  return;
  }
t1=(float)(i)/tx.input_speed;
if(t1 > 1.05 || t1 < 0.95)
  {
  lirerr(231012);
  return;
  }
tx.input_speed=i;
*/
}

void close_tx_input(void)
{
xz("close tx");
if(close(tx_audio_in)==-1)lirerr(231037);  
tx_audio_in=-1;
}



void spproc_setup(void)
{
int i;
init_txmem();
if(kill_all_flag)return;
open_tx_input();
if(kill_all_flag)return;
linrad_thread_create(THREAD_TX_INPUT);
if(kill_all_flag)return;
lir_sleep(50000);
clear_screen();
i=0;
while(!kill_all_flag)
  {
  i++;
  lir_sem_wait(SEM_TXINPUT);
  test_keyboard();
  }
lir_join(THREAD_USER_COMMAND);
close_tx_input();
}

void init_txmem(void)
{
int i;
// Speech processing is done in both the time domain and in the
// frequency domain.
// The lowest frequency of interest is 150 Hz so it will be
// sufficient to use FFTs that span 10 milliseconds.
// With sin squared windows we use 50% overlap and make a read
// every 5 millisecond. (or up to 10 ms for the size to be a power of two)
tx_fftsize=tx.input_speed*.01;
make_power_of_two(&tx_fftsize);
tx_mask=tx_fftsize-1;
tx_read_bytes=tx_fftsize*tx.input_channels*tx.input_bytes/2;
tx_input_block=tx.input_channels*tx_fftsize/2;
// Allow the input buffer to hold 320 milliseconds of data (or more)
tx_micbuf_size=64*tx_input_block;
tx_micbuf_mask=tx_micbuf_size-1;
init_memalloc(txmem, MAX_TXMEM_ARRAYS);
if(tx.input_bytes == 2)
  {
  mem(1,&tx_micbuf_shi,tx_micbuf_size*sizeof(short int),0);
  }
else
  {
  mem(1,&tx_micbuf_int,tx_micbuf_size*sizeof(int),0);
  }
tx_rfbuf_size=8*tx_fftsize;
tx_rfbuf_mask=tx_rfbuf_size-1;
mem(12,&tx_fft1,tx_rfbuf_size*sizeof(float),0);
mem(2,&tx_rf1,tx_rfbuf_size*sizeof(float),0);
mem(12,&tx_fft2,tx_rfbuf_size*sizeof(float),0);
mem(3,&tx_rf2,tx_rfbuf_size*sizeof(float),0);
// Allow 0.32 seconds in the output buffer (or more).
tx_output_size=tx.output_speed*0.32;
make_power_of_two(&tx_output_size);
tx_output_mask=tx_output_size-1;
if(tx.output_bytes ==2)
  {
  mem(4,&tx_outbuf_shi,tx_output_size*sizeof(short int),0);
  }
else
  {
  mem(4,&tx_outbuf_int,tx_output_size*sizeof(int),0);
  }
mem(5,&tx_tmp,2*tx_fftsize*sizeof(float),0);
mem(6,&tx_fft_table,tx_fftsize*sizeof(COSIN_TABLE),0);
mem(7,&tx_fft_permute,2*tx_fftsize*sizeof(short int),0);
mem(8,&tx_fft_win,tx_fftsize*sizeof(float),0);
txmem_size=memalloc((int*)(&txmem_handle),"txmem");
if(txmem_size == 0)
  {
  lirerr(1261);
  return;
  }
tx_fftn=0;
i=tx_fftsize;
while(i>1)
  {
  i>>=1;
  tx_fftn++;
  }
make_permute(2, tx_fftn, tx_fftsize, tx_fft_permute);
make_window(2,tx_fftsize, 2, tx_fft_win);
make_sincos(2, tx_fftsize, tx_fft_table); 
mic_pa=0;
mic_px=0;
}


void tx_setup(void)
{
/*
int dev_flag[MAX_DEVNAMES];
int dev_max_speed[MAX_DEVNAMES];
int dev_min_speed[MAX_DEVNAMES];
int rdwr_channels;  
int rdwr_speed;
int line, column;
int device_no;
audio_buf_info ad_info;
int rdwr_fmt;
int nn, m;
unsigned int device_id, no_of_dadev;

*/
unsigned int device_id;
char s[256];
int line;
unsigned no_of_txda;
int maxline, maxcolumn, rdbytes;
int i, j, k;
int *wgi;
FILE *txpar_file;
char *testbuff;
MMRESULT mmrs;
WAVEOUTCAPS pwoc;

testbuff = malloc(4096);
if(testbuff == NULL)
  {
  lirerr(1044);
  return;
  }
wgi=(void*)(&tx.input_speed);
rx_daout_channels=ui.min_output_channels;
genparm[DA_OUTPUT_SPEED]=ui.min_output_speed;
genparm[SECOND_FFT_ENABLE]=0;
fft1_size=32768;
tx_daout_block=256;
rx_daout_block=256;
timf1_sampling_speed=11025;
txpar_file=fopen("par_tx","r");
  {
  if(txpar_file != NULL)
    {
    rdbytes=fread(testbuff,1,4095,txpar_file);
    fclose(txpar_file);
    if(rdbytes >= 4095)goto do_tx_setup;
    k=0;
    for(i=0; i<MAX_TXPARM; i++)
      {
      while( (testbuff[k]==' ' || testbuff[k]== '\n' ) && k<rdbytes)k++;
      j=0;
      while(testbuff[k]== txparm_text[i][j] && k<rdbytes)
        {
        k++;
        j++;
        } 
      if(txparm_text[i][j] != 0)goto do_tx_setup;    
      while(testbuff[k]!='[' && k<rdbytes)k++;
      if(k>=rdbytes)goto do_tx_setup;
      sscanf(&testbuff[k],"[%d]",&wgi[i]);
      while(testbuff[k]!='\n' && k<rdbytes)k++;
      if(k>=rdbytes)goto do_tx_setup;
      }
    }
  } 
if(tx.check!=TXPAR_VERNR ||
   tx.ad_device_no < 0 ||
   tx.ad_device_no >= MAX_DEVNAMES ||
   tx.da_device_no < 0 ||
   tx.da_device_no >= MAX_DEVNAMES ||
   tx.output_channels < 1 ||
   tx.output_channels > 2 ||
   (tx.output_bytes !=2 && tx.output_bytes !=4) ||
   (tx.input_bytes !=2 && tx.input_bytes !=4) )goto do_tx_setup;
// *******************************************************
// ***   The A/D and A/D setup is already done.
// ***   Ask the user what he wants.
// ***   The transmit functions of Linrad are in a very early stage
// ***   and there is not much to choose from yet............
// ******************************************************
clear_screen();
lir_text(5,5,"A=Speech processor lab");
lir_text(5,6,"B=Soundcard setup for Tx.");
lir_text(5,7,"C=Speech processor setup.");
lir_text(5,8,"X=Return to main menu");
kbd_sel:;
await_processed_keyboard();
if(kill_all_flag) goto txend;
switch (lir_inkey)
  {
  case 'A':
  genparm[DA_OUTPUT_SPEED]=tx.input_speed;
  speech_proc_lab();
  genparm[DA_OUTPUT_SPEED]=ui.min_output_speed;
  goto txend;
  
  case 'B':
  goto do_tx_setup;

  case 'C':
  spproc_setup();
  goto txend;

  
  case 'X':
  goto txend;
  
  default:
  goto kbd_sel;
  }
do_tx_setup:;   
clear_screen();
settextcolor(15);
lir_text(25,2,"TX mode set-up");
rx_daout_bytes=ui.max_output_bytes;
rx_daout_channels=ui.max_output_channels;
if(ui.ad_device_no < SPECIFIC_DEVICE_CODES)open_rx_sndin();
if(kill_all_flag) goto txend;
open_rx_sndout();
if(kill_all_flag) goto txend;
if(ui.ad_device_no < SPECIFIC_DEVICE_CODES)
  {
  read(rx_audio_in, testbuff, 1024);
  if(ui.ad_device_no > 255) read(rx_audio_in2, testbuff, 1024);
  }
for(i=0;i<1024;i++)testbuff[i]=0;
write(rx_audio_out, testbuff, 1024);
// First step through all devices and check which ones
// we can open in write mode while the rx devices are open.
maxline=screen_last_line-1;
maxcolumn=screen_last_col;
settextcolor(14);
lir_text(10,4,"Available soundcards for tx output");
settextcolor(7);
no_of_txda=waveInGetNumDevs();
if(no_of_txda == 0)
  {
  lir_text(5,2,"No soundcard detected (while rx active).");
  lir_text(5,3,press_any_key);
  await_keyboard();
  return;
  }
line=5;
if(no_of_txda > MAX_WAV_DEVICES)no_of_txda=MAX_WAV_DEVICES;
for(device_id=0; device_id<no_of_txda; device_id++)
  {
  mmrs=waveOutGetDevCaps(device_id, &pwoc, sizeof(WAVEOUTCAPS));   
  if(mmrs==MMSYSERR_NOERROR)
    {
    sprintf(s,"%d  %s", device_id, pwoc.szPname);
    lir_text(10,line,s);
    line++;
    }
  }
close_rx_sndout();
close_rx_sndin();
line+=2;  
if(no_of_txda == 1)
  {
  tx.ad_device_no=0;
  }
else  
  {
  lir_text(0,line,"Select device for Tx output by line number:");
  tx.ad_device_no=lir_get_integer(51,line, 2, 0, no_of_txda-1);
  if(kill_all_flag) return;
  line++;
  }
// Windows does not allow us to query the device for its capabilities.
// Ask the user to specify what he wants.
win_query_warning(&line);
lir_text(10,line,"Sampling speed (Hz):");
tx.output_speed=lir_get_integer(31,line, 6, 5000, 999999);
if(kill_all_flag) return;
get_rxbits:;
lir_text(40,line,"No of bits (16/24):");
i=lir_get_integer(60,line, 2, 16, 24);
if(kill_all_flag) return;
if(i==16)
  {
  tx.output_bytes=2;
  }
else
  {
  if(i==24)  
    {
    tx.output_bytes=4;
    }
  else
    {  
    goto get_rxbits;
    }
  }    
line+=2;
lir_text(10,line,"Select Tx output mode:");
line++;
lir_text(10,line,"1: One channel (real data)");
line++;
lir_text(10,line,"2: Two channels (I/Q complex data)");
line+=2;
lir_text(10,line,"=>");
tx.output_channels=lir_get_integer(13,line, 1, 1, 2);
clear_screen();
// Look for a device for microphone input.
if(ui.ad_device_no < SPECIFIC_DEVICE_CODES)open_rx_sndin();
if(kill_all_flag) goto txend;
open_rx_sndout();
if(kill_all_flag) goto txend;
open_tx_output();
if(kill_all_flag) goto txend;
if(ui.ad_device_no < SPECIFIC_DEVICE_CODES)
  {
  read(rx_audio_in, testbuff, 1024);
  if(ui.ad_device_no > 255) read(rx_audio_in2, testbuff, 1024);
  }
for(i=0;i<1024;i++)testbuff[i]=0;
/*
write(audio_out, testbuff, 1024);
write(tx_audio_out, testbuff, 1024);
// First step through all devices and check which ones
// we can open in read mode while the rx devices and tx output are open.
settextcolor(14);
lir_text(10,4,"Checking device drivers for microphone input");
settextcolor(7);

nn=0;
line=5;


for(device_no=0; device_no<MAX_DEVNAMES; device_no++)
  {
  if(ioctl(rx_audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
    {
    lirerr(281028);
    return;
    }
  if(ui.ad_device_no < SPECIFIC_DEVICE_CODES)
    {
    while(ad_info.bytes > 1024)
      {
      read(rx_audio_in, testbuff, 1024);
      if(ioctl(rx_audio_in, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
        {
        lirerr(291028);
        return;
        }
      }
    if(ui.ad_device_no > 255)
      {
      if(ioctl(rx_audio_in2, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
        {
        lirerr(311028);
        return;
        }
      while(ad_info.bytes > 1024)  
        {
        read(rx_audio_in2, testbuff, 1024);
        if(ioctl(rx_audio_in2, SNDCTL_DSP_GETISPACE, &ad_info) == -1)
          {
          lirerr(321028);
          return;
          }
        }  
      }
    }
  for(i=0;i<1024;i++)testbuff[i]=0;
  if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
    {
    lirerr(331028);
    return;
    }
  while(ad_info.bytes > 1024)
    {
    write(audio_out, testbuff, 1024);
    if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
      {
      lirerr(341028);
      return;
      }
    }
  if(ioctl(tx_audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
    {
    lirerr(351028);
    return;
    }
  while(ad_info.bytes > 1024)
    {
    write(tx_audio_out, testbuff, 1024);
    if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &ad_info) == -1)
      {
      lirerr(361028);
      return;
      }
    }
  rdwr_channels=0;
  rdwr_fmt=0;
  if(device_no != rx_audio_in)
    {
    make_devname(device_no);
      {
      sprintf(s,"Trying to open %s %s. Device defective if system hangs here.",
                                           dev_name, devmode_txt[0]);
      settextcolor(12);
      lir_text(0,line+1,s);
      settextcolor(7);
      tx_audio_in=open( dev_name, O_RDONLY|O_NONBLOCK, 0);
      if(tx_audio_in != -1)
        {
        line++;
        if(line >= maxline)
          {
          for(column=0; column<maxcolumn; column++)s[column]=' ';
          s[maxcolumn]=0;
          lir_text(0,line,s);
          settextcolor(14);
          lir_text(10,line,"(press ENTER to continue)");
          settextcolor(7);
          await_keyboard();
          if(kill_all_flag) goto txend;
          lir_text(0,line,s);
          line=1;
          }
// Check for 16 bits
        if(ioctl(tx_audio_in, SNDCTL_DSP_GETFMTS, &m) == -1)
          {
mic_baddev:;
          rdwr_channels=0;
          goto mic_cls;
          }
        rdwr_fmt=0;  
        if( (m&AFMT_S16_LE) != 0)
          {
          rdwr_fmt=16;
          }
#ifdef AFMT_S32_LE
// Check for 32 bits
        if( (m&AFMT_S32_LE) != 0)
          {
          rdwr_fmt+=32;
          }
#endif   
        if(rdwr_fmt == 0)goto mic_cls;
// Find max no of channels.
// We never want more than 2 tx output channels!!
        k=2;
        rdwr_channels=0;
mic_find_channels:;
        m=k;
        if(ioctl(tx_audio_in, SNDCTL_DSP_CHANNELS, &m) == -1 || k != m)
          {
          k/=2;
          if(k == 0)
            {
            goto mic_baddev;
            }
          goto mic_find_channels;
          }
        rdwr_channels|=k;
        k/=2;
        if(k != 0)goto mic_find_channels;
// Find maximum output speed
        rdwr_speed=ABOVE_MAX_SPEED;
        if(ioctl(tx_audio_in, SNDCTL_DSP_SPEED, &rdwr_speed)==-1)
          {
          goto mic_baddev;
          }
        dev_max_speed[device_no]=rdwr_speed;
// Find minimum output speed
        rdwr_speed=5000;
        if(ioctl(tx_audio_in, SNDCTL_DSP_SPEED, &rdwr_speed)==-1)
          {
          goto mic_baddev;
          }
        dev_min_speed[device_no]=rdwr_speed;
        settextcolor(7);  
        sprintf(s,"  %02d  %s     ",device_no,dev_name);
        lir_text(0,line,s);
        sprintf(s,"%s   ",devmode_txt[0]);
        if(rdwr_channels==3)
          {
          strcat(s,"1,2");
          }
        else
          {
          sprintf(testbuff,"%d",rdwr_channels);
          strcat(s,testbuff);
          }
        strcat(s," channels      "); 
        if(rdwr_fmt == 48)
          {
          strcat(s,"16,32");
          }
        else
          {
          sprintf(testbuff,"%d",rdwr_fmt);
          strcat(s,testbuff);
          }
        strcat(s," bits      "); 
        if(rdwr_speed != dev_max_speed[device_no])
          {
          sprintf(testbuff,"%d to %d",rdwr_speed,dev_max_speed[device_no]);
          strcat(s,testbuff);
          }
        else
          {
          sprintf(testbuff,"%d",rdwr_speed);
          strcat(s,testbuff);
          }
        strcat(s," Hz         ");    
        lir_text(DEVNAME_SIZE+5,line,s);
        nn++;
mic_cls:;
        if(close(tx_audio_in)==-1)
          {
          lirerr(371037);  
          return;
          }
        }
      }          
    }      
  dev_flag[device_no]=rdwr_fmt+rdwr_channels;
  }
close_rx_sndout();
close_rx_sndin();
close_tx_output();
for(i=0; i<80; i++)s[i]=' ';
s[80]=0;
line++;
lir_text(0,line,s);
line++;
if(nn == 0)
  {
  lir_text(0,line,"No device found for microphone input.");
  lir_text(2,line+1,"ERROR (press any key)");
  await_keyboard();
  goto txend;
  }
settextcolor(14);
lir_text(0,line,"Select device for microphone input");
get_micdev:;
settextcolor(7);
tx.ad_device_no=lir_get_integer(37, line, 3, 0,MAX_DEVNAMES-1);
if( (3 & dev_flag[tx.ad_device_no]) == 0)
  {
  settextcolor(15);
  lir_text(29,line,"???");
  lir_text(25,line+1,"ERROR (press any key)");
  await_keyboard();
  if(kill_all_flag) goto txend;
  clear_lines(line+1,line+1);
  goto get_micdev;
  }
// Use one channel if the device can do both one and two channels.
if( (3 & dev_flag[tx.ad_device_no]) == 3)
  {
  tx.input_channels=1;
  }
else
  {
  tx.input_channels=dev_flag[tx.ad_device_no];
  }
clear_screen();
if( (48 & dev_flag[tx.ad_device_no]) == 48)
  {
  lir_text(5,10,"Select microphone no of bits");
  lir_text(3,12,"1=16 bits");
  lir_text(3,13,"2=32 bits");
get_micbits:;
  await_keyboard();
  if(kill_all_flag) goto txend;
  if(lir_inkey=='1')
    {
    tx.input_bytes=2;
    }
  else
    {  
    if(lir_inkey=='2')
      {
      tx.input_bytes=4;
      }
    else
      {
      goto get_micbits;
      }
    }    
  }
else
  {
  tx.input_bytes=(16+32)&dev_flag[tx.ad_device_no];
  tx.input_bytes>>=3;
  }
if(dev_min_speed[tx.ad_device_no] != dev_max_speed[tx.ad_device_no])
  {
  clear_screen();
  sprintf(s,"Select microphone A/D speed %d to %d",
               dev_min_speed[tx.ad_device_no], dev_max_speed[tx.ad_device_no]);
  lir_text(5,11,s);
  i=0;
  while(s[i] != 0)i++;
  tx.input_speed=lir_get_integer(i+7, 11, 6, dev_min_speed[tx.ad_device_no],
                                          dev_max_speed[tx.ad_device_no]);
  }
else
  {  
  tx.input_speed=dev_min_speed[tx.ad_device_no];
  }
clear_screen();
make_devname(tx.ad_device_no);

sprintf(s,"Microphone input: %s, %d channel(s), %d bits, %d Hz",dev_name,
                      tx.input_channels,tx.input_bytes<<3,tx.input_speed);
settextcolor(14);
lir_text(15,0,s);

txpar_file=fopen("par_tx","w");
if(txpar_file == NULL)
  {
  lirerr(381164);
  return;
  }
tx.check=TXPAR_VERNR;
for(i=0; i<MAX_TXPARM; i++)
  {
  fprintf(txpar_file,"%s [%d]\n",txparm_text[i],wgi[i]);
  }
parfile_end(txpar_file);
await_keyboard();
*/
txend:;
free(testbuff);
settextcolor(7);
lir_inkey=0;
}
